import torch
from tqdm import tqdm
import torch.nn.functional as F
import pandas as pd
import numpy as np
import time
from .Net import DePassAE, DePassAE_va
from sklearn.neighbors import kneighbors_graph
from .utils import *
from torch_geometric.loader import ClusterData, ClusterLoader
from sklearn.decomposition import PCA
from scipy.sparse import coo_matrix
from torch.optim.lr_scheduler import MultiStepLR

class DePass:
    def __init__(
                 self,
                 data: dict,
                 data_type: str = 'spatial',
                 device: torch.device = torch.device('cpu'),
                 learning_rate: float = 2e-4,
                 epochs: int = 200,
                 batch_training: bool = False,
                 sub_graph_size: int = 1024,
                 updata_step: int = 25,
                 mlpLayer1: list = [256, 64],
                 mlpLayer2: list = [256, 64],
                 mlpLayer3: list = [256, 64],
                 dim: int = 64,
                 K_spatial: int = 5,
                 K_feature: int = 20,
                 ):
        
        r"""
        Initialize the DePass model.

        Args:
            data: dict: Input data dictionary containing multiple modality data.
            data_type: str='spatial': Data type, which can be 'spatial' or 'single_cell'.
            device: torch.device=torch.device('cpu'): Computing device.
            learning_rate: float=2e-4: Learning rate.
            epochs: int=200: Total number of training epochs.
            batch_training: bool=False: Whether to use batch training.
            sub_graph_size: int=1024: sub graph size for batch training.
            updata_step: int=25: Update step.
            mlpLayer1: list=[256, 64]: Structure of the first MLP layer.
            mlpLayer2: list=[256, 64]: Structure of the second MLP layer.
            mlpLayer3: list=[256, 64]: Structure of the third MLP layer.
            dim: int=64: Embedding dimension.
            K_spatial: int=5: Number of neighbors for the spatial KNN graph.
            K_feature: int=20: Number of neighbors for the feature KNN graph.
        """
        self.data = data.copy()
        self.data_type = data_type
        self.device = device
        self.learning_rate = learning_rate
        self.epochs = epochs
        self.batch_training = batch_training
        self.sub_graph_size = sub_graph_size
        self.updata_step = updata_step
        self.mlpLayer1 = mlpLayer1
        self.mlpLayer2 = mlpLayer2
        self.mlpLayer3 = mlpLayer3
        self.dim = dim
        self.K_spatial = K_spatial
        self.K_feature = K_feature

        self.n_views = len(self.data)

        # if self.n_views not in [2, 3]:
        #     raise ValueError("The number of input modalities is more than 3. We suggest integrating 2 or 3 at a time, but got {}".format(self.n_views))
       
        if self.data_type not in ['spatial', 'single_cell']:
            raise ValueError("The data type must be 'spatial' or 'single_cell', but got {}".format(self.data_type))
        
        first_key = next(iter(self.data.keys()))
        if self.data_type == 'spatial':
            if 'spatial' not in self.data[first_key].obsm:
                raise ValueError("The 'spatial' key is not found in `adata.obsm`. Please provide the spatial coordinates.")
            
        if self.device.type == 'cuda':
           device_name = torch.cuda.get_device_name(0)
        else:
           device_name ='CPU'

        print('[Config]')
        print('Modalities:',self.n_views,'| Data:',self.data_type,'| Device:',device_name,'\n')   

    def train(self):
        if self.n_views == 2:
            modality = list(self.data.keys())
            self.mlpLayer1=[128,64] if modality[0] in ['protein'] else [256,64]
            self.mlpLayer2=[128,64] if modality[1] in ['protein'] else [256,64]
            self.adata_omics1 = self.data[modality[0]]
            self.adata_omics2 = self.data[modality[1]]

            self.features_omics1 = torch.FloatTensor(self.adata_omics1.obsm['input_feat'].copy()).to(self.device)
            self.features_omics2 = torch.FloatTensor(self.adata_omics2.obsm['input_feat'].copy()).to(self.device)

            self.dim_input1 = self.features_omics1.shape[1]
            self.dim_input2 = self.features_omics2.shape[1]
           
            if self.features_omics1.shape[0] > 4e4:
                self.train2_sparse() # Use sparse format and batch training
            else:
                self.train2() 

        if self.n_views >= 3:
            modality = list(self.data.keys())
            self.mlpLayer_list = []
            self.dim_input_list = []
            self.adata_list = []
            self.features_list = []
            
            for m in modality:
                mlpLayer = [128, 64] if m in ['protein'] else [256, 64]
                self.mlpLayer_list.append(mlpLayer)
                
                adata_omics = self.data[m]
                self.adata_list.append(adata_omics)
                
                features_omics = torch.FloatTensor(adata_omics.obsm['input_feat'].copy()).to(self.device)
                self.features_list.append(features_omics)
                
                dim_input = features_omics.shape[1]
                self.dim_input_list.append(dim_input)
    
            self.train_va()
     

    def train2(self):
        self.model = DePassAE(dim_input1=self.dim_input1, dim_input2=self.dim_input2, mlplayer1=self.mlpLayer1,
                           mlplayer2=self.mlpLayer2,dim=self.dim).to(self.device)

        self.optimizer = torch.optim.Adam(self.model.parameters(),lr=self.learning_rate)
        self.scheduler = MultiStepLR(self.optimizer, milestones=[80,160], gamma=0.5)
        
        print('[Initializing]')
        print('Graph Construction : Running...')
        adj1 = construct_knn_graph_hnsw(data=self.adata_omics1.obsm['input_feat'].copy(),k=self.K_feature)
        adj2 = construct_knn_graph_hnsw(data=self.adata_omics2.obsm['input_feat'].copy(),k=self.K_feature)
       
        if self.data_type == 'spatial':
            xy = self.adata_omics1.obsm['spatial'].copy()
            adj_shared = construct_knn_graph_hnsw(data=xy,k=self.K_spatial)

        elif self.data_type == 'single_cell':
            adj_shared = adj1 + adj2
            adj_shared.data[adj_shared.data > 1] = 1
        print('Graph Construction : Done!')

        print('Data Enhancement : Running...')
        k_xNeigh = 3
        k_zNeigh = 10 if self.data_type == 'spatial' else 3
        FeatNeigh1 = construct_knn_graph_hnsw(data=self.adata_omics1.obsm['input_feat'].copy(),k=k_xNeigh)
        FeatNeigh2 = construct_knn_graph_hnsw(data=self.adata_omics2.obsm['input_feat'].copy(),k=k_xNeigh)
        SpatialNeigh = adj_shared if self.data_type == 'spatial' else None
        x1_self_enh = get_augment(data=self.features_omics1,SpatialNeigh=SpatialNeigh,FeatNeigh=FeatNeigh1,w_fea=0.1,w_spa=0.1)            
        x2_self_enh = get_augment(data=self.features_omics2,SpatialNeigh=SpatialNeigh,FeatNeigh=FeatNeigh2,w_fea=0.1,w_spa=0.1) 
        print('Data Enhancement : Done!')

        x1_z_enh =  x1_self_enh
        x2_z_enh =  x2_self_enh

        adj1_norm = normalize_adj(adj1).to(self.device)
        adj2_norm = normalize_adj(adj2).to(self.device)
        adj_shared_norm = normalize_adj(adj_shared).to(self.device)
    
        if self.sub_graph_size:
            print('')
            undirected_geo = to_undirected_geo_data(adj_shared,node_index=torch.arange(adj_shared.shape[0]))
            cluster_data = ClusterData(undirected_geo, num_parts=int(np.ceil(undirected_geo.num_nodes / self.sub_graph_size)) * 4,
                                       recursive=False)
            train_loader = ClusterLoader(cluster_data, batch_size=2, shuffle=True)
        
        print('')
        print('[Training]')
        print("Model training starts...")  
        self.model.train()
        for epoch in tqdm(range(0, self.epochs)):
            e1, e2 = self.model.mlp_forward(x1=x1_self_enh, z1=x1_z_enh, x2=x2_self_enh, z2=x2_z_enh)
            if self.batch_training:
                loss_recon1_list = []
                loss_recon2_list = []
                for cur in train_loader:
                    np_index = cur.node_index.numpy()
                    e1_batch = e1[np_index]
                    e2_batch = e2[np_index]
                    adj1_batch = adj1[np_index, :][:, np_index]
                    adj2_batch = adj2[np_index, :][:, np_index]
                    adj_shared_batch = adj_shared[np_index, :][:, np_index]
                    adj1_batch = normalize_adj(adj1_batch).to(self.device)
                    adj2_batch = normalize_adj(adj2_batch).to(self.device)
                    adj_shared_batch = normalize_adj(adj_shared_batch).to(self.device)
                    results = self.model.gcn_forward(e1_batch, e2_batch, adj_shared_batch, adj1_batch, adj2_batch)
                    loss_recon1_ = F.mse_loss(self.features_omics1[np_index], results['emb_recon1'])
                    loss_recon2_ = F.mse_loss(self.features_omics2[np_index], results['emb_recon2'])
                    loss_recon1_list.append(loss_recon1_)
                    loss_recon2_list.append(loss_recon2_)
                loss_recon1 = torch.mean(torch.stack(loss_recon1_list))
                loss_recon2 = torch.mean(torch.stack(loss_recon2_list))
            else:    
                results = self.model.gcn_forward(e1, e2, adj_shared_norm, adj1_norm, adj2_norm)
                loss_recon1 = F.mse_loss(self.features_omics1, results['emb_recon1'])
                loss_recon2 = F.mse_loss(self.features_omics2, results['emb_recon2'])

            loss = loss_recon1 + loss_recon2 
         
            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()
            self.scheduler.step()  

            if epoch > 0 and epoch % self.updata_step == 0:
                with torch.no_grad():
                    self.model.eval()
                    results = self.model.gcn_inference(e1,e2,adj_shared_norm,adj1_norm,adj2_norm)
                    self.model.train()
                    ZNeigh = construct_knn_graph_hnsw(data=results['z'].detach().cpu().numpy(),k=k_zNeigh)
                    x1_z_enh = get_augment(data=self.features_omics1,SpatialNeigh=SpatialNeigh,FeatNeigh=ZNeigh)            
                    x2_z_enh = get_augment(data=self.features_omics2,SpatialNeigh=SpatialNeigh,FeatNeigh=ZNeigh) 
                    adj1 = construct_knn_graph_hnsw(data=e1.detach().cpu().numpy(),k=self.K_feature)
                    adj2 = construct_knn_graph_hnsw(data=e2.detach().cpu().numpy(),k=self.K_feature)
                    adj1_norm = normalize_adj(adj1).to(self.device)
                    adj2_norm = normalize_adj(adj2).to(self.device)
                    if self.data_type == 'single_cell':
                        adj_shared = adj1 + adj2
                        adj_shared.data[adj_shared.data > 1] = 1
                        adj_shared_norm = normalize_adj(adj_shared).to(self.device)

        
        print("Model training finished!\n")
        with torch.no_grad():
            self.model.eval()
            results = self.model.gcn_inference(e1, e2, adj_shared_norm, adj1_norm, adj2_norm)
            ZNeigh = construct_knn_graph_hnsw(data=results['z'].detach().cpu().numpy(),k=k_zNeigh)
            x1_z_enh = get_augment(data=self.features_omics1,SpatialNeigh=SpatialNeigh,FeatNeigh=ZNeigh)             
            x2_z_enh = get_augment(data=self.features_omics2,SpatialNeigh=SpatialNeigh,FeatNeigh=ZNeigh)  
      
        self.embedding = norm_convert(results['z'])
        self.z1 = norm_convert(results['z1'])
        self.z2 = norm_convert(results['z2'])
        self.alpha = results['alpha'].detach().cpu().numpy()
        self.x1_enh = x1_z_enh.detach().cpu().numpy()
        self.x2_enh = x2_z_enh.detach().cpu().numpy()

    def train2_sparse(self):
        self.model = DePassAE(dim_input1=self.dim_input1, dim_input2=self.dim_input2, mlplayer1=self.mlpLayer1,
                           mlplayer2=self.mlpLayer2,dim=self.dim).to(self.device)
        self.optimizer = torch.optim.Adam(self.model.parameters(),lr=self.learning_rate)
        self.scheduler = MultiStepLR(self.optimizer, milestones=[80,160], gamma=0.5)

        print('[Initializing]')
        print('Graph Construction : Running...')
        adj1 = construct_knn_graph_hnsw_sparse(self.adata_omics1.obsm['input_feat'].copy(), k=self.K_feature)
        adj2 = construct_knn_graph_hnsw_sparse(self.adata_omics2.obsm['input_feat'].copy(), k=self.K_feature)
        if self.data_type == 'spatial':
            adj_shared = construct_knn_graph_hnsw_sparse(self.adata_omics1.obsm['spatial'].copy(), k=self.K_spatial)
        elif self.data_type == 'single_cell':
            adj_shared = adj1 + adj2
            adj_shared.data[adj_shared.data > 1] = 1
        print('Graph Construction : Done!')

        print('Data Enhancement : Running...')
        k_xNeigh = 6
        # k_zNeigh = int(self.K_feature/2) if self.data_type == 'spatial' else 6
        k_zNeigh = 20 if self.data_type == 'spatial' else 6
        FeatNeigh1 = construct_knn_graph_hnsw_sparse(self.adata_omics1.obsm['input_feat'].copy(), k=k_xNeigh)
        FeatNeigh2 = construct_knn_graph_hnsw_sparse(self.adata_omics2.obsm['input_feat'].copy(), k=k_xNeigh)
        SpatialNeigh = adj_shared if self.data_type == 'spatial' else None
        x1_self_enh = augment_data_lc(self.features_omics1, SpatialNeigh=SpatialNeigh,FeatNeigh=FeatNeigh1,w_fea=0.1,w_spa=0.1)
        x2_self_enh = augment_data_lc(self.features_omics2, SpatialNeigh=SpatialNeigh,FeatNeigh=FeatNeigh2,w_fea=0.1,w_spa=0.1)
        x1_z_enh = x1_self_enh 
        x2_z_enh = x2_self_enh 
        print('Data Enhancement : Done!')

        adj1_norm = normalize_adj_scr(adj1).to(self.device)
        adj2_norm = normalize_adj_scr(adj2).to(self.device)
        adj_shared_norm = normalize_adj_scr(adj_shared).to(self.device)

        # batch_training
        undirected_geo = to_undirected_geo_data(adj_shared,node_index=torch.arange(adj_shared.shape[0]))
        cluster_data = ClusterData(undirected_geo, num_parts=int(np.ceil(undirected_geo.num_nodes / self.sub_graph_size)) * 4,
                                       recursive=False)
        train_loader = ClusterLoader(cluster_data, batch_size=2, shuffle=True)

        print('')
        print('[Training]')
        print("Model training starts...")
        self.model.train()
        for epoch in tqdm(range(0, self.epochs)):
            e1, e2 = self.model.mlp_forward(x1=x1_self_enh, z1=x1_z_enh, x2=x2_self_enh, z2=x2_z_enh)
            loss_recon1_list = []
            loss_recon2_list = []
            for cur in train_loader:
                np_index = cur.node_index.numpy()
                e1_batch = e1[np_index]
                e2_batch = e2[np_index]
                adj1_batch = sparse_indexing(adj1, np_index)
                adj2_batch = sparse_indexing(adj2, np_index)
                adj_shared_batch = sparse_indexing(adj_shared, np_index)
                adj1_batch = normalize_adj_scr(adj1_batch).to(self.device)
                adj2_batch = normalize_adj_scr(adj2_batch).to(self.device)
                adj_shared_batch = normalize_adj_scr(adj_shared_batch).to(self.device)
                results = self.model.gcn_forward(e1_batch, e2_batch, adj_shared_batch, adj1_batch, adj2_batch)
                loss_recon1_ = F.mse_loss(self.features_omics1[np_index], results['emb_recon1'])
                loss_recon2_ = F.mse_loss(self.features_omics2[np_index], results['emb_recon2'])
                loss_recon1_list.append(loss_recon1_)
                loss_recon2_list.append(loss_recon2_)

            loss_recon1 = torch.mean(torch.stack(loss_recon1_list))
            loss_recon2 = torch.mean(torch.stack(loss_recon2_list))
            loss = loss_recon1 + loss_recon2 
            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()
            self.scheduler.step()

            if epoch>0 and epoch % self.updata_step == 0:
                with torch.no_grad():
                    self.model.eval()
                    results = self.model.gcn_inference(e1,e2,adj_shared_norm,adj1_norm,adj2_norm)
                    self.model.train()
                    ZNeigh = construct_knn_graph_hnsw_sparse(results['z'].detach().cpu().numpy(), k=k_zNeigh) 
                    x1_z_enh = augment_data_lc(self.features_omics1,SpatialNeigh=SpatialNeigh,FeatNeigh=ZNeigh)   
                    x2_z_enh = augment_data_lc(self.features_omics2,SpatialNeigh=SpatialNeigh,FeatNeigh=ZNeigh) 
                    adj1 = construct_knn_graph_hnsw_sparse(e1.detach().cpu().numpy(), k=self.K_feature)
                    adj2 = construct_knn_graph_hnsw_sparse(e2.detach().cpu().numpy(), k=self.K_feature)
                    adj1_norm = normalize_adj_scr(adj1).to(self.device)
                    adj2_norm = normalize_adj_scr(adj2).to(self.device)
                    
                    if self.data_type == 'single_cell':
                        adj_shared = adj1 + adj2
                        adj_shared.data[adj_shared.data > 1] = 1
                        adj_shared_norm = normalize_adj_scr(adj_shared).to(self.device)

        print("Model training finished!\n")
        with torch.no_grad():
            self.model.eval()
            results = self.model.gcn_inference(e1, e2, adj_shared_norm, adj1_norm, adj2_norm)
            ZNeigh = construct_knn_graph_hnsw_sparse(results['z'].detach().cpu().numpy(), k=k_zNeigh) 
            x1_z_enh  = augment_data_lc(self.features_omics1,SpatialNeigh=SpatialNeigh,FeatNeigh=ZNeigh)   
            x2_z_enh  = augment_data_lc(self.features_omics2,SpatialNeigh=SpatialNeigh,FeatNeigh=ZNeigh) 
     
        self.embedding = norm_convert(results['z'])
        self.z1 = norm_convert(results['z1'])
        self.z2 = norm_convert(results['z2'])
        self.alpha = results['alpha'].detach().cpu().numpy()
        self.x1_enh  = x1_z_enh.detach().cpu().numpy()
        self.x2_enh  = x2_z_enh.detach().cpu().numpy()
    
    def train_va(self):
        self.model = DePassAE_va(dim_input_list=self.dim_input_list, mlplayer_list=self.mlpLayer_list, dim=self.dim).to(self.device)
        self.optimizer = torch.optim.Adam(self.model.parameters(), lr=self.learning_rate)
        self.scheduler = MultiStepLR(self.optimizer, milestones=[80, 160], gamma=0.5)

        print('[Initializing]')
        print('Graph Construction : Running...')
        adj_list = []
        for adata in self.adata_list:
            adj = construct_knn_graph_hnsw(data=adata.obsm['input_feat'].copy(), k=self.K_feature)
            adj_list.append(adj)

        if self.data_type == 'spatial':
            xy = self.adata_list[0].obsm['spatial'].copy()
            adj_shared = construct_knn_graph_hnsw(data=xy, k=self.K_spatial)
        elif self.data_type == 'single_cell':
            adj_shared = sum(adj_list)
            adj_shared.data[adj_shared.data > 1] = 1
        print('Graph Construction : Done!')

        print('Data Enhancement : Running...')
        k_xNeigh = 3
        k_zNeigh = int(self.K_feature/2) if self.data_type == 'spatial' else 3
        FeatNeigh_list = []
        for adata in self.adata_list:
            FeatNeigh = construct_knn_graph_hnsw(data=adata.obsm['input_feat'].copy(), k=k_xNeigh)
            FeatNeigh_list.append(FeatNeigh)

        SpatialNeigh = adj_shared if self.data_type == 'spatial' else None
        x_self_enh_list = []
        for i in range(len(self.features_list)):
            x_self_enh = get_augment(data=self.features_list[i], SpatialNeigh=SpatialNeigh, FeatNeigh=FeatNeigh_list[i], w_fea=0.1, w_spa=0.1)
            x_self_enh_list.append(x_self_enh)
        x_z_enh_list =  x_self_enh_list 

        print('Data Enhancement : Done!')

        adj_norm_list = []
        for adj in adj_list:
            adj_norm = normalize_adj(adj).to(self.device)
            adj_norm_list.append(adj_norm)
        adj_shared_norm = normalize_adj(adj_shared).to(self.device)

        print('')
        print('[Training]')
        print("Model training starts...")
        self.model.train()
        for epoch in tqdm(range(0, self.epochs)):
            e_list = self.model.mlp_forward(xs=x_self_enh_list, zs=x_z_enh_list)
            results = self.model.gcn_forward(e_list, adj_shared_norm, adj_norm_list)
            loss_recon_list = []
            for i in range(len(self.features_list)):
                loss_recon = F.mse_loss(self.features_list[i], results['emb_recons'][i])
                loss_recon_list.append(loss_recon)

            loss = sum(loss_recon_list)
            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()
            self.scheduler.step()

            if epoch > 0 and epoch % self.updata_step == 0:
                with torch.no_grad():
                    self.model.eval()
                    results = self.model.gcn_inference(e_list, adj_shared_norm, adj_norm_list)
                    self.model.train()
                    ZNeigh = construct_knn_graph_hnsw(data=results['z'].detach().cpu().numpy(), k=k_zNeigh)

                    x_z_enh_list = []
                    for i in range(len(self.features_list)):
                        x_z_enh = get_augment(data=self.features_list[i], SpatialNeigh=SpatialNeigh, FeatNeigh=ZNeigh)
                        x_z_enh_list.append(x_z_enh)

                    for i in range(len(e_list)):
                        adj = construct_knn_graph_hnsw(data=e_list[i].detach().cpu().numpy(), k=self.K_feature)
                        adj_norm_list[i] = normalize_adj(adj).to(self.device)

                    if self.data_type == 'single_cell':
                        adj_shared = sum(adj_list)
                        adj_shared.data[adj_shared.data > 1] = 1
                        adj_shared_norm = normalize_adj(adj_shared).to(self.device)

        print("Model training finished!\n")
        with torch.no_grad():
            self.model.eval()
            results = self.model.gcn_inference(e_list, adj_shared_norm, adj_norm_list)
            ZNeigh = construct_knn_graph_hnsw(data=results['z'].detach().cpu().numpy(), k=k_zNeigh)
            x_z_enh_list = []
            for i in range(len(self.features_list)):
                x_z_enh = get_augment(data=self.features_list[i], SpatialNeigh=SpatialNeigh, FeatNeigh=ZNeigh)
                x_z_enh_list.append(x_z_enh)

        self.embedding = norm_convert(results['z'])
        self.alpha = results['alpha'].detach().cpu().numpy()
        for i, x in enumerate(x_z_enh_list):
            setattr(self, f'x{i+1}_enh', x.detach().cpu().numpy())
    


import torch
from scipy.sparse import coo_matrix

def get_augment(data=None,FeatNeigh=None, SpatialNeigh=None, w = 0.01, w_fea=0.25,w_spa=0.25):
        FeatNeigh = FeatNeigh.to(data.device)
        if SpatialNeigh is None:
            enhanced_data = data + w * FeatNeigh @ data 
        if SpatialNeigh is not None:
            SpatialNeigh = SpatialNeigh.to(data.device)
            enhanced_data = data + w_fea * FeatNeigh @ data + w_spa * SpatialNeigh @ data      
        return enhanced_data

def toTensor_sparse(data,device):
        num_samples, _ = data.shape
        if not isinstance(data, coo_matrix):
            data = data.tocoo()
        indices = torch.tensor([data.row, data.col], dtype=torch.long, device=device)
        values = torch.tensor(data.data, dtype=torch.float32, device=device)
        data = torch.sparse.FloatTensor(indices, values, (num_samples, num_samples))
        return data

def augment_data_lc(data, FeatNeigh=None, SpatialNeigh=None, w=0.01, w_fea=0.25, w_spa=0.25):
    FeatNeigh = FeatNeigh.to(data.device)
    if SpatialNeigh is None:
                enhanced_data = data + w * FeatNeigh @ data 
        
    if SpatialNeigh is not None:
       SpatialNeigh = SpatialNeigh.to(data.device)
       enhanced_data = data.clone()
      
       feat_contribution = w_fea * torch.sparse.mm(FeatNeigh, data)
       enhanced_data += feat_contribution
       del feat_contribution  
       if torch.cuda.is_available():
           torch.cuda.empty_cache()  
       
       spa_contribution = w_spa * torch.sparse.mm(SpatialNeigh, data)
       enhanced_data += spa_contribution
       del spa_contribution  
       if torch.cuda.is_available():
            torch.cuda.empty_cache()
    
    return enhanced_data

def norm_convert(emb):
    emb = F.normalize(emb, p=2, eps=1e-12, dim=1)
    return emb.detach().cpu().numpy()



import numpy as np
import scipy
import os
import random
import torch
from torch.backends import cudnn
import anndata
import scanpy as sc
import torch_geometric

############# Graph Construction #############

def construct_knn_graph_hnsw(data: np.ndarray, k: int = 20, space: str = r'l2') -> torch.Tensor:
    r"""
    Construct a k-NN graph using the HNSW algorithm.

    Parameters:
        data (numpy.ndarray): Input data matrix (num_samples x dim).
        k (int): Number of nearest neighbors (default: 20).
        space (str): Distance metric (r'l2' for Euclidean distance, default).

    Returns:
        adj (torch.Tensor): Adjacency matrix of the k-NN graph (dense tensor).
    """

    import hnswlib
    from torch_geometric.utils import to_dense_adj
    data = data.astype(np.float32)
    num_samples, dim = data.shape
    p = hnswlib.Index(space=space, dim=dim)
    p.init_index(max_elements=num_samples, ef_construction=200, M=16)
    p.add_items(data)
    p.set_ef(50)
    indices, _ = p.knn_query(data, k=k)
    row_indices = np.repeat(np.arange(num_samples), k)
    col_indices = indices.flatten()
    edge_index = torch.tensor([row_indices, col_indices], dtype=torch.long)
    adj = to_dense_adj(edge_index)[0]
    return adj


import numpy as np
import torch
import hnswlib
from typing import Union

def construct_knn_graph_hnsw_sparse(
    data: np.ndarray,
    k: int = 20,
    space: str = 'l2',
    symmetric: bool = False,
    sparse_format: str = 'torch_coo'
) -> Union[torch.Tensor, scipy.sparse.spmatrix]:
    r"""
    Construct a k-NN graph using HNSW and return a sparse adjacency matrix.

    Parameters:
        data (np.ndarray): Input data matrix (num_samples x dim).
        k (int): Number of nearest neighbors.
        space (str): Distance metric ('l2' or 'ip').
        symmetric (bool): Whether to make the adjacency matrix symmetric (undirected graph).
        sparse_format (str): Output format ('coo', 'csr', 'torch_coo').

    Returns:
        adj (Union[torch.Tensor, scipy.sparse.spmatrix]): Sparse adjacency matrix.
    """
    # --- Step 1: Generate HNSW k-NN indices ---
    data = data.astype(np.float32)
    num_samples, dim = data.shape

    # Build HNSW index
    p = hnswlib.Index(space=space, dim=dim)
    p.init_index(max_elements=num_samples, ef_construction=200, M=16)
    p.add_items(data)
    p.set_ef(50)
    
    # Query k-NN
    indices, _ = p.knn_query(data, k=k)
    
    # --- Step 2: Generate edge indices ---
    row_indices = np.repeat(np.arange(num_samples), k)
    col_indices = indices.flatten()
    
    # Remove self-loops
    mask = row_indices != col_indices
    row_indices = row_indices[mask]
    col_indices = col_indices[mask]

    # --- Step 3: Symmetrize edges (if needed) ---
    if symmetric:
        # Add reverse edges
        row_indices_sym = np.concatenate([row_indices, col_indices])
        col_indices_sym = np.concatenate([col_indices, row_indices])
        
        # Remove duplicates
        edges = np.vstack([row_indices_sym, col_indices_sym]).T
        edges = np.unique(edges, axis=0)
        row_indices, col_indices = edges[:, 0], edges[:, 1]

    # --- Step 4: Build sparse adjacency matrix ---
    if sparse_format in ['coo', 'csr']:
        # SciPy sparse matrix
        import scipy.sparse as sp
        adj = sp.coo_matrix(
            np.ones_like(row_indices), 
            (row_indices, col_indices),
            shape=(num_samples, num_samples)
        )
        if sparse_format == 'csr':
            adj = adj.tocsr()
    
    elif sparse_format == 'torch_coo':
        # PyTorch sparse tensor (COO format)
        indices = torch.vstack([
            torch.LongTensor(row_indices),
            torch.LongTensor(col_indices)
        ])
        adj = torch.sparse_coo_tensor(
            indices, 
            torch.ones(indices.shape[1]), 
            size=(num_samples, num_samples)
        )
    
    else:
        raise ValueError(f"Invalid sparse_format: {sparse_format}")

    return adj

def to_undirected_geo_data(adj_shared, node_index=None) -> torch_geometric.data.Data:
    r"""
    Converts the input adjacency matrix (sparse matrix, sparse tensor, dense tensor, or dense NumPy array)
    into an undirected graph's PyTorch Geometric Data object.

    Parameters:
        adj_shared: The input adjacency matrix (sparse matrix, sparse tensor, dense tensor, or dense NumPy array).
        node_index (optional): Node indices.

    Returns:
        undirected_geo (torch_geometric.data.Data): The converted undirected graph Data object.
    """
    from torch_geometric.data import Data
    from torch_geometric.transforms import ToUndirected
    from torch_geometric.utils import dense_to_sparse, from_scipy_sparse_matrix
    import scipy.sparse as sp

    if sp.issparse(adj_shared):
        edge_index, edge_attr = from_scipy_sparse_matrix(adj_shared)
    elif adj_shared.is_sparse:
        edge_index = adj_shared.coalesce().indices()
        edge_attr = adj_shared.coalesce().values()
    elif isinstance(adj_shared, np.ndarray):
        adj_shared = torch.tensor(adj_shared, dtype=torch.float)
        edge_index, edge_attr = dense_to_sparse(adj_shared)
    else:
        edge_index, edge_attr = dense_to_sparse(adj_shared)

    node_index = data2input(node_index) if node_index is not None else None
    geo_dataset = Data(edge_index=edge_index, edge_attr=edge_attr, node_index=node_index)
    transform = ToUndirected()
    undirected_geo = transform(geo_dataset)
    return undirected_geo

def data2input(data):
    import torch
    import scipy.sparse as sp
    import numpy as np
    if sp.issparse(data):
        data = data.toarray()
    if not isinstance(data, torch.Tensor):
        data = torch.LongTensor(data) if str(data.dtype).startswith("int") else torch.FloatTensor(data)
    return data

import torch
import numpy as np
from scipy.sparse import coo_matrix

import torch

def sparse_indexing(adj, np_index):
    """
    Indexing a subset of rows and columns from a PyTorch sparse tensor without using for loops.

    Parameters:
        adj (torch.sparse.FloatTensor): Input adjacency matrix in PyTorch sparse format.
        np_index (np.ndarray): Indices of the rows and columns to keep.

    Returns:
        torch.sparse.FloatTensor: Submatrix corresponding to the indices.
    """
    # --- Step 1: Ensure adj is in COO format ---
    adj = adj.coalesce()
    edge_index = adj.indices()
    edge_weight = adj.values()

    # --- Step 2: Create a mask for the edges to keep ---
    row_indices, col_indices = edge_index

    # Convert np_index to a tensor for faster lookups
    np_index_tensor = torch.tensor(np_index, dtype=torch.long)

    # Create the mask using torch.isin for rows and columns
    row_mask = torch.isin(row_indices, np_index_tensor)
    col_mask = torch.isin(col_indices, np_index_tensor)
    mask = row_mask & col_mask

    # --- Step 3: Extract sub-indices and sub-values ---
    sub_row_indices = row_indices[mask]
    sub_col_indices = col_indices[mask]
    sub_values = edge_weight[mask]

    # --- Step 4: Create index map ---
    # Create a tensor that maps old indices to new indices
    index_map = torch.zeros(torch.max(np_index_tensor) + 1, dtype=torch.long)
    index_map[np_index_tensor] = torch.arange(len(np_index_tensor))

    # --- Step 5: Map row and column indices to the new indexing ---
    sub_row_indices = index_map[sub_row_indices]
    sub_col_indices = index_map[sub_col_indices]

    # --- Step 6: Create the sub-adjacency matrix ---
    sub_shape = (len(np_index), len(np_index))
    sub_adj = torch.sparse.FloatTensor(
        torch.stack([sub_row_indices, sub_col_indices]),
        sub_values,
        size=sub_shape
    ).coalesce()

    return sub_adj


########## Data Preprocessing ##########

def clr_normalize_each_cell(adata: anndata.AnnData, inplace: bool = True) -> anndata.AnnData:
    r"""
    Normalize count vector for each cell, i.e., for each row of .X.
    """

    def seurat_clr(x):
        r"""
        CLR normalization function (Seurat implementation).
        """
        s = np.sum(np.log1p(x[x > 0]))
        exp = np.exp(s / len(x))
        return np.log1p(x / exp)

    if not inplace:
        adata = adata.copy()

    adata.X = np.apply_along_axis(
        seurat_clr, 1, (adata.X.A if scipy.sparse.issparse(adata.X) else np.array(adata.X))
    )
    return adata


def lsi(adata: anndata.AnnData, n_components: int = 20, use_highly_variable: bool = None, **kwargs) -> None:
    r"""
    Perform LSI analysis (following the Seurat v3 approach).
    """

    import sklearn.preprocessing
    import sklearn.utils.extmath

    if use_highly_variable is None:
        use_highly_variable = "highly_variable" in adata.var

    adata_use = adata[:, adata.var["highly_variable"]] if use_highly_variable else adata

    X = tfidf(adata_use.X)

    X_norm = sklearn.preprocessing.Normalizer(norm="l1").fit_transform(X)
    X_norm = np.log1p(X_norm * 1e4)

    X_lsi = sklearn.utils.extmath.randomized_svd(X_norm, n_components, **kwargs)[0]
    X_lsi -= X_lsi.mean(axis=1, keepdims=True)
    X_lsi /= X_lsi.std(axis=1, ddof=1, keepdims=True)

    adata.obsm["X_lsi"] = X_lsi[:, 1:]


def tfidf(X) -> np.ndarray:
    r"""
    Perform TF-IDF normalization (following the Seurat v3 approach).
    """

    import scipy

    idf = X.shape[0] / X.sum(axis=0)

    if scipy.sparse.issparse(X):
        tf = X.multiply(1 / X.sum(axis=1))
        return tf.multiply(idf)
    else:
        tf = X / X.sum(axis=1, keepdims=True)
        return tf * idf
    
def fix_seed(seed: int, deterministic_cudnn: bool = True, set_hash_seed: bool = True, mode: str = r'') -> None:
    r"""
    Fix random seed for reproducibility.

    Parameters:
        seed (int): Random seed value.
        deterministic_cudnn (bool): Whether to enable deterministic CUDA operations (default: True).
        set_hash_seed (bool): Whether to set the hash seed (default: True).
        mode (str): Mode for strictness (default: r'').
    """
   
    random.seed(seed)
    np.random.seed(seed)

    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed) 
        torch.cuda.manual_seed_all(seed)
        if deterministic_cudnn:
            torch.backends.cudnn.deterministic = True
            torch.backends.cudnn.benchmark = False
    
    if set_hash_seed:
        os.environ['PYTHONHASHSEED'] = str(seed)

    # strict mode
    if mode == r'strict':
        ### This will slow down the learning process
        if hasattr(torch, 'use_deterministic_algorithms'):
            torch.use_deterministic_algorithms(True)
        os.environ['CUBLAS_WORKSPACE_CONFIG'] = r':16:8'

def filter_genes_retain(adata, min_cells, retain_genes):
    default_filter = sc.pp.filter_genes(adata, min_cells=min_cells, inplace=False)
    # default_filter = adata.var["n_cells"] >= min_cells

    retain_mask = adata.var_names.isin(retain_genes)
    combined_filter = default_filter | retain_mask

    return adata[:, combined_filter].copy()

import numpy as np
from anndata import AnnData
from typing import Optional, List

def filter_genes_keep_list(
    adata: AnnData,
    gene_list: List[str],
    min_cells: Optional[float] = None,
    copy: bool = False
) -> Optional[AnnData]:
   
    if min_cells is None:
        min_cells_threshold = adata.shape[0] * 0.01  
    else:
        min_cells_threshold = min_cells  

    valid_genes = [gene for gene in gene_list if gene in adata.var_names]
    
    n_cells_per_gene = np.asarray((adata.X > 0).sum(axis=0)).flatten()

    mask_min_cells = n_cells_per_gene >= min_cells_threshold
    mask_keep_genes = adata.var_names.isin(valid_genes)
    final_mask = mask_min_cells | mask_keep_genes

    if copy:
        return adata[:, final_mask].copy()
    else:
        adata._inplace_subset_var(final_mask)

def preprocess_data(adata: anndata.AnnData, modality: str, n_lsi: int = 64, n_top_genes:int=1000, filter_adt: int = 0, gene_list: list = None) -> None:
    r"""
    Preprocess data based on modality.

    Parameters:
        adata (anndata.AnnData): Input AnnData object.
        modality (str): Data modality (e.g., 'rna', 'protein', 'atac').
        n_lsi (int): Number of LSI components (default: 64).
        filter_adt (int): Filter threshold for ADT data (default: 0).
    """
    adata.var_names_make_unique()

    if modality == 'atac':
        sc.pp.filter_genes(adata, min_cells=adata.shape[0] * 0.01)
        lsi(adata, use_highly_variable=False, n_components=n_lsi)

    elif modality == 'rna':
        if gene_list is not None:
            filter_genes_keep_list(adata, gene_list=gene_list, min_cells=0.01)
        else:
            sc.pp.filter_genes(adata, min_cells=adata.shape[0] * 0.01)
        sc.pp.highly_variable_genes(adata, flavor="seurat_v3", n_top_genes=n_top_genes)
        sc.pp.normalize_total(adata, target_sum=1e4)
        sc.pp.log1p(adata)
        sc.pp.scale(adata)

        if gene_list is not None:
            highly_var_mask = adata.var['highly_variable']
            marker_mask = adata.var_names.isin(gene_list)
            hvg = highly_var_mask | marker_mask
        else:
            hvg = adata.var['highly_variable']
        adata.var['highly_variable_all'] =hvg
        adata.obsm['X_norm'] = adata[:, hvg].X

    elif modality == 'protein':
        if filter_adt !=0: 
            sc.pp.filter_genes(adata, min_cells=filter_adt)
        adata = clr_normalize_each_cell(adata)
        sc.pp.scale(adata)
        adata.obsm['X_clr'] = adata.X

    elif modality == 'metabolite':
        sc.pp.filter_genes(adata, min_cells=adata.shape[0] * 0.01)
        sc.pp.highly_variable_genes(adata, flavor="seurat_v3", n_top_genes=n_top_genes)
        sc.pp.normalize_total(adata, target_sum=1e4)
        sc.pp.log1p(adata)
        sc.pp.scale(adata)
        adata.obsm['X_norm'] = adata[:, adata.var['highly_variable']].X

import torch
from torch_geometric.utils import add_self_loops, degree

def normalize_adj_scr(adj: torch.sparse.FloatTensor, add_loop=True) -> torch.sparse.FloatTensor:
    """
    Normalize adjacency matrix (symmetric normalization: D^{-1/2} A D^{-1/2}).

    Parameters:
        adj (torch.sparse.FloatTensor): Input adjacency matrix in COO sparse format.
        add_loop (bool): Whether to add self-loops to the adjacency matrix. Default is True.

    Returns:
        torch.sparse.FloatTensor: Normalized adjacency matrix in COO sparse format.
    """
    # --- Step 1: Extract edge indices and values from sparse tensor ---
    adj = adj.coalesce()
    edge_index = adj.indices()
    edge_weight = adj.values()
    num_nodes = adj.size(0)

    # --- Step 2: Add self-loops (optional) ---
    if add_loop:
        edge_index, _ = add_self_loops(edge_index, num_nodes=num_nodes)

    # --- Step 3: Compute degrees and normalization factors ---
    row, col = edge_index
    deg = degree(col, num_nodes=num_nodes, dtype=torch.float32)
    deg_inv_sqrt = deg.pow(-0.5)
    deg_inv_sqrt[deg_inv_sqrt == float('inf')] = 0  # Handle isolated nodes

    # --- Step 4: Apply symmetric normalization ---
    norm = deg_inv_sqrt[row] * deg_inv_sqrt[col]

    # --- Step 5: Build normalized sparse tensor ---
    normalized_adj = torch.sparse.FloatTensor(
        edge_index,
        norm,
        size=(num_nodes, num_nodes)
    ).coalesce()

    return normalized_adj
 

def normalize_adj(adj_matrix: torch.Tensor) -> torch.sparse.FloatTensor:
    r"""
    Normalize adjacency matrix for dense or sparse tensor format.

    Parameters:
        adj_matrix (torch.Tensor): Input adjacency matrix (dense or sparse tensor).

    Returns:
        torch.sparse.FloatTensor: Normalized adjacency matrix in sparse tensor format.
    """
    num_nodes = adj_matrix.size(0)
    adj_matrix = adj_matrix + torch.eye(num_nodes)
    adj_matrix = (adj_matrix + adj_matrix.t()) / 2
    deg = torch.sum(adj_matrix, dim=1)
    deg_inv_sqrt = torch.pow(deg, -0.5)
    deg_inv_sqrt[deg_inv_sqrt == float('inf')] = 0
    adj_norm = deg_inv_sqrt.view(-1, 1) * adj_matrix * deg_inv_sqrt.view(1, -1)
    indices = adj_norm.nonzero(as_tuple=False)
    values = adj_norm[indices[:, 0], indices[:, 1]]
    sparse_adj = torch.sparse.FloatTensor(indices.t(), values, adj_matrix.size())
    return sparse_adj


############# Clustering #############

def mclust_R(adata: anndata.AnnData, num_cluster: int, modelNames: str = r'EEE', used_obsm: str = r'emb_pca', random_seed: int = 2024) -> anndata.AnnData:
    r"""
    Clustering using the mclust algorithm.
    The parameters are the same as those in the R package mclust.

    Parameters:
        adata (anndata.AnnData): Input AnnData object.
        num_cluster (int): Number of clusters.
        modelNames (str): Model name for mclust (default: r'EEE').
        used_obsm (str): Key for the embedding in adata.obsm (default: r'emb_pca').
        random_seed (int): Random seed for reproducibility (default: 2024).

    Returns:
        adata (anndata.AnnData): AnnData object with clustering results stored in adata.obs['mclust'].
    """
    import rpy2.robjects as robjects
    from rpy2.robjects import numpy2ri
    numpy2ri.activate()

    np.random.seed(random_seed)
    robjects.r.library("mclust")
    r_random_seed = robjects.r['set.seed']
    r_random_seed(random_seed)
    rmclust = robjects.r['Mclust']

    res = rmclust(numpy2ri.numpy2rpy(adata.obsm[used_obsm]), num_cluster, modelNames)
    mclust_res = np.array(res[-2])

    adata.obs['mclust'] = mclust_res.astype('int')
    adata.obs['mclust'] = adata.obs['mclust'].astype('category')
    return adata


def pca(adata: anndata.AnnData, use_reps: str = None, n_comps: int = 20) -> np.ndarray:
    r"""
    Dimension reduction with PCA algorithm.

    Parameters:
        adata (anndata.AnnData): Input AnnData object.
        use_reps (str): Key for the input representation in adata.obsm (default: None).
        n_comps (int): Number of PCA components (default: 20).

    Returns:
        feat_pca (numpy.ndarray): PCA-transformed features.
    """
    from sklearn.decomposition import PCA
    from scipy.sparse import csc_matrix, csr_matrix

    pca_model = PCA(n_components=n_comps)
    if use_reps is not None:
        feat_pca = pca_model.fit_transform(adata.obsm[use_reps])
    else:
        if isinstance(adata.X, (csc_matrix, csr_matrix)):
            feat_pca = pca_model.fit_transform(adata.X.toarray())
        else:
            feat_pca = pca_model.fit_transform(adata.X)
    return feat_pca


import torch.nn.functional as F

def norm(array: np.ndarray) -> np.ndarray:
    r"""
    Normalize an array using PyTorch's normalize function.

    Parameters:
        array (np.ndarray): Input array to be normalized.

    Returns:
        np.ndarray: Normalized array.
    """
    import torch
    tensor = torch.tensor(array, dtype=torch.float32)
    normalized_tensor = F.normalize(tensor, p=2, eps=1e-12, dim=1)
    normalized_array = normalized_tensor.detach().numpy()
    return normalized_array


def clustering(adata: anndata.AnnData, n_clusters: int = 7, key: str = r'emb', add_key: str = None, method: str = r'mclust', 
               start: float = 0.05, end: float = 3.0, increment: float = 0.1, use_pca: bool = False, n_comps: int = 20, use_X: bool = False) -> None:
    r"""
    Perform spatial clustering based on latent representations.

    Parameters:
        adata (anndata.AnnData): Input AnnData object containing spatial data.
        n_clusters (int): Target number of clusters (default: 7).
        key (str): Key for accessing embeddings in adata.obsm (default: r'emb').
        add_key (str): Key to store clustering results in adata.obs (default: None).
        method (str): Clustering algorithm (r'mclust', 'leiden', 'louvain', or 'kmeans') (default: r'mclust').
        start (float): Starting resolution for search (leiden/louvain only) (default: 0.05).
        end (float): Maximum resolution for search (leiden/louvain only) (default: 3.0).
        increment (float): Resolution search step size (default: 0.1).
        use_pca (bool): Whether to apply PCA dimensionality reduction (default: False).
        n_comps (int): Number of PCA components if use_pca=True (default: 20).
        use_X (bool): Whether to use adata.X as input features (default: False).
    """
    from sklearn.cluster import KMeans
    import scanpy as sc

    if use_X:
        adata.obsm['X'] = adata.X  
        key = 'X' 

    if use_pca:
        if adata.obsm[key].shape[1] > n_comps:
            adata.obsm[key + '_pca'] = pca(adata, use_reps=key, n_comps=n_comps)
        else:
            use_pca = False  

    if method == 'mclust':
        rep_key = key + '_pca' if use_pca else key
        adata = mclust_R(adata, used_obsm=rep_key, num_cluster=n_clusters)
        if add_key:
            adata.obs[add_key] = adata.obs['mclust']

    elif method == 'leiden':
        rep_key = key + '_pca' if use_pca else key
        res = binary_search_res(
            adata, 
            n_clusters, 
            use_rep=rep_key, 
            method=method, 
            res_min=start,
            res_max=end
        )
        sc.tl.leiden(adata, random_state=2024, resolution=res)
        adata.obs['leiden'] = adata.obs['leiden'].astype('category')
        if add_key:
            adata.obs[add_key] = adata.obs['leiden']

    elif method == 'kmeans':
        rep_key = key + '_pca' if use_pca else key
        kmeans = KMeans(n_clusters=n_clusters, random_state=2024).fit(adata.obsm[rep_key])
        adata.obs['kmeans'] = kmeans.labels_
        adata.obs['kmeans'] = adata.obs['kmeans'].astype('category')
        if add_key:
            adata.obs[add_key] = adata.obs['kmeans']


def binary_search_res(adata: anndata.AnnData, n_clusters: int, method: str = 'leiden', use_rep: str = 'emb', 
                     res_min: float = 0.01, res_max: float = 3.0, max_iter: int = 100, tol: float = 1e-2, 
                     random_state: int = 2024) -> float:
    r"""
    Binary search to find optimal resolution parameter for target cluster number.

    Parameters:
        adata (anndata.AnnData): Input AnnData object.
        n_clusters (int): Target number of clusters.
        method (str): Clustering method ('leiden' or 'louvain').
        use_rep (str): Key for embeddings in adata.obsm.
        res_min (float): Minimum resolution to search.
        res_max (float): Maximum resolution to search.
        max_iter (int): Maximum search iterations.
        tol (float): Tolerance for matching target cluster count.
        random_state (int): Random seed for reproducibility.

    Returns:
        float: Optimal resolution parameter.

    Raises:
        ValueError: If target cluster count cannot be achieved.
    """

    if 'neighbors' not in adata.uns:
        sc.pp.neighbors(adata, n_neighbors=50, use_rep=use_rep)

    for i in range(max_iter):
        current_res = (res_min + res_max) / 2
        print(f"Iteration {i+1}: Testing resolution = {current_res:.4f}")

        if method == 'leiden':
            sc.tl.leiden(adata, resolution=current_res, random_state=random_state)
            n_found = len(adata.obs['leiden'].unique())
        else:
            raise ValueError("Supported methods: 'leiden' or 'louvain'")

        print(f"Resolution {current_res:.4f} produced {n_found} clusters")

        if abs(n_found - n_clusters) <= tol:
            print(f"Optimal resolution found: {current_res:.4f}")
            return current_res
        elif n_found > n_clusters:
            res_max = current_res
        else:
            res_min = current_res

    raise ValueError(f"Failed to find resolution in {max_iter} iterations. " +
                    "Try adjusting res_min/res_max or increasing max_iter.")



################# Evaluation ###################

def super_eval(y_pred, y_true) -> dict:
    r"""
    Evaluates clustering performance using various supervised metrics.
    Handles both numeric and categorical/string labels.

    Args:
        y_pred (array-like): Predicted cluster labels (can be numeric, string or categorical)
        y_true (array-like): True cluster labels (can be numeric, string or categorical)

    Returns:
        dict: Dictionary containing evaluation metrics:
            - AMI: Adjusted Mutual Information
            - NMI: Normalized Mutual Information
            - ARI: Adjusted Rand Index
            - Homogeneity: Homogeneity score
            - V-measure: Harmonic mean of homogeneity and completeness
            - Mutual Information: Raw mutual information

    Raises:
        ValueError: If inputs contain NaN/infinite values or have mismatched lengths
    """
    from sklearn import metrics
    import numpy as np
    import pandas as pd

    def _preprocess_labels(y):
        if isinstance(y, (pd.Categorical, pd.Series)):
            y = y.astype(str) 
        y = np.array(y)
    
        if y.dtype.kind in ['U', 'O']:  
            y = pd.factorize(y)[0]
        return y.astype(float)


    y_pred = _preprocess_labels(y_pred)
    y_true = _preprocess_labels(y_true)

    if len(y_pred) != len(y_true):
        raise ValueError(f"Length mismatch: y_pred ({len(y_pred)}) vs y_true ({len(y_true)})")
    
    if np.isnan(y_true).any():
        raise ValueError("y_true contains NaN values after conversion")
    if np.isnan(y_pred).any():
        raise ValueError("y_pred contains NaN values after conversion")
    if np.isinf(y_true).any():
        raise ValueError("y_true contains infinite values")
    if np.isinf(y_pred).any():
        raise ValueError("y_pred contains infinite values")

    results = {
        'AMI': metrics.adjusted_mutual_info_score(y_true, y_pred),
        'NMI': metrics.normalized_mutual_info_score(y_true, y_pred),
        'ARI': metrics.adjusted_rand_score(y_true, y_pred),
        'Homogeneity': metrics.homogeneity_score(y_true, y_pred),
        'V-measure': metrics.v_measure_score(y_true, y_pred),
        'Mutual Information': metrics.mutual_info_score(y_true, y_pred),
    }

    return results


def unsuper_eval(X, y) -> dict:
    r"""
    Evaluates the performance of a clustering algorithm using various unsupervised metrics.

    Args:
        X (array-like): The feature matrix.
        y (array-like): The predicted cluster labels.

    Returns:
        dict: A dictionary containing the evaluation metrics and their corresponding values.
    """

    from sklearn.metrics import silhouette_score, calinski_harabasz_score, davies_bouldin_score

    if not isinstance(X, np.ndarray):
        X = np.array(X)
    if not isinstance(y, np.ndarray):
        y = np.array(y)

    assert not np.isnan(X).any(), "X contains NaN values"
    assert not np.isnan(y).any(), "y contains NaN values"

    results = {
        'ASW': 0.5 * (silhouette_score(X, y) + 1),
        'DBI': davies_bouldin_score(X, y),
        'CHI': calinski_harabasz_score(X, y),
    }

    return results


import torch
import numpy as np
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.parameter import Parameter
from torch.nn.modules.module import Module
from typing import Any, Optional, Tuple
from torch.autograd import Function
from .layer import *

class DePassAE(Module):
    def __init__(self, dim_input1, dim_input2, mlplayer1, mlplayer2, dim, dropout=0.2, act='elu'):
        super(DePassAE, self).__init__() 

        self.mlplayer1=mlplayer1
        self.mlplayer2=mlplayer2
        self.dim=dim
        self.act = act
        self.dropout = dropout

        self.mlp_encoder1 = mlp_encoder(dim_input1, self.mlplayer1,act=self.act)
        self.mlp_encoder2 = mlp_encoder(dim_input2, self.mlplayer2,act=self.act)

        self.gcn_encoder_s1 = GCN(self.mlplayer1[-1],self.dim)
        self.gcn_encoder_s2 = GCN(self.mlplayer2[-1],self.dim) 
        self.gcn_encoder_con = GCN(self.mlplayer1[-1]+self.mlplayer1[-1],self.dim) 

        # self._gcn_encoder_s1 =  GCN(self.dim,self.dim)
        # self._gcn_encoder_s2 =  GCN(self.dim,self.dim) 
        # self._gcn_encoder_con = GCN(self.dim,self.dim) 

        # self.__gcn_encoder_s1 =  GCN(self.dim,self.dim)
        # self.__gcn_encoder_s2 =  GCN(self.dim,self.dim) 
        # self.__gcn_encoder_con = GCN(self.dim,self.dim) 

        self.Eff_Att1 = EfficientAdditiveAttention(self.dim)
        self.Eff_Att2 = EfficientAdditiveAttention(self.dim)

        self.atten_cross = AttentionLayer(2 * self.dim, 2 * self.dim)
        # self.lin1 = nn.Linear(2 * self.dim,32)
        # self.elu = nn.ELU()
        # self.lin2 = nn.Linear(64,32)
        # self.atten_cross = AttentionLayer(1 * self.dim, 1 * self.dim)

        # self._gcn_decoder1 = GCN(2 * self.dim, 2 * self.dim)
        # self._gcn_decoder2 = GCN(2 * self.dim, 2 * self.dim)
        # self.__gcn_decoder1 = GCN(2 * self.dim, 2 * self.dim)
        # self.__gcn_decoder2 = GCN(2 * self.dim, 2 * self.dim)

        self.gcn_decoder1 = GCN(2 * self.dim, dim_input1)
        self.gcn_decoder2 = GCN(2 * self.dim, dim_input2)

        # self.gcn_decoder1 = GCN(32, dim_input1)
        # self.gcn_decoder2 = GCN(32, dim_input2)

    def mlp_forward(self,x1,z1,x2,z2):
        x1 = self.mlp_encoder1(x1)
        z1 = self.mlp_encoder1(z1)
        x2 = self.mlp_encoder2(x2)
        z2 = self.mlp_encoder2(z2)

        e1 = torch.mean(torch.stack([x1, z1], dim=0), dim=0)
        e2 = torch.mean(torch.stack([x2, z2], dim=0), dim=0)

        return e1, e2
    
    def gcn_forward(self, e1_batch, e2_batch, adj_shared_batch, adj1_batch, adj2_batch):
       
        E_fusion = torch.cat([e1_batch, e2_batch], dim=-1)
        s1 = self.gcn_encoder_s1(e1_batch, adj1_batch)
        s2 = self.gcn_encoder_s2(e2_batch, adj2_batch)
        f =  self.gcn_encoder_con(E_fusion, adj_shared_batch)

        # s1 = self._gcn_encoder_s1(s1, adj1_batch)
        # s2 = self._gcn_encoder_s2(s2, adj2_batch)
        # f =  self._gcn_encoder_con(f, adj_shared_batch)

        # s1 = self.__gcn_encoder_s1(s1, adj1_batch)
        # s2 = self.__gcn_encoder_s2(s2, adj2_batch)
        # f =  self.__gcn_encoder_con(f, adj_shared_batch)

        s1 = self.Eff_Att1(f,s1) 
        s2 = self.Eff_Att2(f,s2) 
     

        z1 = torch.cat([s1, f], dim=-1)
        z2 = torch.cat([s2, f], dim=-1)

        # z1=s1
        # z2=s2
        z, alpha_omics_1_2 = self.atten_cross(z1, z2)
        # z = self.lin1(z)
        # z = self.elu(z)
        # z = self.lin2(z)

        # emb_recon1 = self._gcn_decoder1(z, adj_shared_batch)
        # emb_recon2 = self._gcn_decoder2(z, adj_shared_batch)

        # emb_recon1 = self.__gcn_decoder1(z, adj_shared_batch)
        # emb_recon2 = self.__gcn_decoder2(z, adj_shared_batch)

        emb_recon1 = self.gcn_decoder1(z, adj_shared_batch)
        emb_recon2 = self.gcn_decoder2(z, adj_shared_batch)

        # emb_reconz1 = self.gcn_decoder1(z1, adj_shared_batch)
        # emb_reconz2 = self.gcn_decoder2(z2, adj_shared_batch)

        results = {
            'z1': z1,
            'z2': z2,
            'z': z,
            'emb_recon1': emb_recon1,
            'emb_recon2': emb_recon2,
            'alpha': alpha_omics_1_2,
        }

        return results

    def gcn_inference(self, e1, e2, adj_shared, adj1, adj2):
        E_fusion = torch.cat([e1, e2], dim=-1)
        
        s1 = self.gcn_encoder_s1(e1, adj1)
        s2 = self.gcn_encoder_s2(e2, adj2)
        f = self.gcn_encoder_con(E_fusion, adj_shared)

        s1 = self.Eff_Att1(f,s1) 
        s2 = self.Eff_Att2(f,s2) 
        # s1 = ScaledDotProductAttention(f,s1) 
        # s2 = ScaledDotProductAttention(f,s2) 

        z1 = torch.cat([s1, f], dim=-1)
        z2 = torch.cat([s2, f], dim=-1)
        # z1=s1
        # z2=s2

        z, alpha_omics_1_2 = self.atten_cross(z1, z2)

        results = {
            'z1': z1,
            'z2': z2,
            'z': z,
            'alpha': alpha_omics_1_2,
        }

        return results

class DePassAE1(Module):
    def __init__(self, dim_input1, mlplayer1, dim, dropout=0.2, act='elu'):
        super(DePassAE1, self).__init__() 
        self.mlplayer1=mlplayer1
        self.dim=dim
        self.act = act
        self.dropout = dropout
        self.mlp_encoder1 = mlp_encoder(dim_input1, self.mlplayer1,act=self.act)
        self.gcn_encoder_s1 = GCN(self.mlplayer1[-1],self.dim)
        self.gcn_encoder_con = GCN(self.mlplayer1[-1],self.dim) 
        self.Eff_Att1 = EfficientAdditiveAttention(self.dim)
        self.gcn_decoder1 = GCN(2 * self.dim, dim_input1)

    def mlp_forward(self,x1,z1):
        x1 = self.mlp_encoder1(x1)
        z1 = self.mlp_encoder1(z1)
        e1 = torch.mean(torch.stack([x1, z1], dim=0), dim=0)
        return e1
    
    def gcn_forward(self, e1_batch, adj_shared_batch, adj1_batch):
        s1 = self.gcn_encoder_s1(e1_batch, adj1_batch)
        f =  self.gcn_encoder_con(e1_batch, adj_shared_batch)
        s1 = self.Eff_Att1(f,s1)
        z = torch.cat([s1, f], dim=-1) 
        emb_recon1 = self.gcn_decoder1(z, adj_shared_batch)
        results = {
            'z': z,
            'emb_recon1': emb_recon1
        }
        return results

    def gcn_inference(self, e1, adj_shared, adj1):
        s1 = self.gcn_encoder_s1(e1, adj1)
        f = self.gcn_encoder_con(e1, adj_shared)
        s1 = self.Eff_Att1(f,s1) 
        z = torch.cat([s1, f], dim=-1)

        results = {
            'z': z
        }

        return results

class DePassAE_va(Module):
    def __init__(self, dim_input_list, mlplayer_list, dim, dropout=0.2, act='elu'):
        super(DePassAE_va, self).__init__()
        self.num_modalities = len(dim_input_list)
        self.mlplayer_list = mlplayer_list
        self.dim = dim
        self.act = act
        self.dropout = dropout

        self.mlp_encoders = nn.ModuleList([
            mlp_encoder(dim_input_list[i], mlplayer_list[i], act=self.act)
            for i in range(self.num_modalities)
        ])

        self.gcn_encoders_s = nn.ModuleList([
            GCN(mlplayer_list[i][-1], self.dim)
            for i in range(self.num_modalities)
        ])

        fused_dim = sum(mlplayer_list[i][-1] for i in range(self.num_modalities))
        self.gcn_encoder_con = GCN(fused_dim, self.dim)

        self.eff_atts = nn.ModuleList([
            EfficientAdditiveAttention(self.dim) for _ in range(self.num_modalities)
        ])

        self.atten_cross = AttentionLayer(2 * self.dim, 2 * self.dim)

        self.gcn_decoders = nn.ModuleList([
            GCN(2 * self.dim, dim_input_list[i]) for i in range(self.num_modalities)
        ])

    def mlp_forward(self,xs,zs):
        es = []  
        for i in range(self.num_modalities):
            x = self.mlp_encoders[i](xs[i])
            z = self.mlp_encoders[i](zs[i])
            e = torch.mean(torch.stack([x, z], dim=0), dim=0)
            es.append(e)
        return es

    def gcn_forward(self, es, adj_shared, adjs):
   
        E_fusion = torch.cat(es, dim=-1)

        ss = []
        for i in range(self.num_modalities):
            s = self.gcn_encoders_s[i](es[i], adjs[i])
            ss.append(s)

        f = self.gcn_encoder_con(E_fusion, adj_shared)

        for i in range(self.num_modalities):
            ss[i] = self.eff_atts[i](f, ss[i])

        fused_features = []
        for i in range(self.num_modalities):
            fused_feature = torch.cat([ss[i], f], dim=-1)
            fused_features.append(fused_feature)

        z, alpha = self.atten_cross(*fused_features)


        emb_recons = []
        for i in range(self.num_modalities):
            emb_recon = self.gcn_decoders[i](z, adj_shared)
            emb_recons.append(emb_recon)

        results = {
            'z': z,
            'alpha': alpha,
            'emb_recons': emb_recons,
        }

        return results

    def gcn_inference(self, es, adj_shared, adjs):

        E_fusion = torch.cat(es, dim=-1)

        ss = []
        for i in range(self.num_modalities):
            s = self.gcn_encoders_s[i](es[i], adjs[i])
            ss.append(s)

        f = self.gcn_encoder_con(E_fusion, adj_shared)

        for i in range(self.num_modalities):
            ss[i] = self.eff_atts[i](f, ss[i])

        fused_features = []
        for i in range(self.num_modalities):
            fused_feature = torch.cat([ss[i], f], dim=-1)
            fused_features.append(fused_feature)

        z, alpha = self.atten_cross(*fused_features)

        results = {
            'z': z,
            'alpha': alpha,
        }

        for i in range(self.num_modalities):
            results[f'z{i+1}'] = ss[i]

        return results


from pathlib import Path
from typing import Optional, Union, Tuple, List
import matplotlib.pyplot as plt
from anndata import AnnData
import scanpy as sc
from typing import Optional


def plot_spatial(  
    adata: AnnData,
    color: str='DePass',  
    save_path: Optional[Union[str, Path]] = None,
    save_name: str = 'spatial_plot',
    title: Optional[str] = None,
    s: int = 35,
    figsize: Tuple[float, float] = (3, 3),
    dpi: int = 300,
    format: str = "pdf",
    frameon: bool = False,
    adjust_margins: bool = True,
    legend_loc: Optional[str] = 'right margin',
    colorbar_loc: Optional[str] = None, 
    show: bool = False,
    **kwargs
) -> None:
    r"""
    Plot spatial data with specified parameters and save the figure.

    Parameters:
        adata (AnnData): Input AnnData object containing spatial data.
        color (str): Column name in adata.obs for coloring the plot.
        save_path (Optional[str]): Path to save the figure (default: None).
        save_name (str): Name of the saved figure.
        title (Optional[str]): Title of the plot (default: None).
        s (int): Size of the markers (default: 35).
        figsize (tuple): Figure size (default: (3, 3)).
        dpi (int): DPI of the saved figure (default: 300).
        format (str): Format of the saved figure (default: "pdf").
        frameon (bool): Whether to show the frame (default: False).
        adjust_margins (bool): Whether to adjust margins (default: True).
        legend_loc (Optional[str]): Location of the legend (default: 'right margin').
        colorbar_loc (Optional[str]): Colorbar position. Set to None to disable colorbar.
        show (bool): Whether to show the plot (default: False).
        **kwargs: Additional keyword arguments for sc.pl.embedding.

    Returns:
        None
    """

    if not isinstance(adata, AnnData):
        raise TypeError("Expected AnnData object, got {}".format(type(adata)))
    
    if not save_name.strip():
        raise ValueError("save_name must contain non-whitespace characters")
    
    file_format = format.lower().lstrip('.')
    allowed_formats = {'png', 'pdf', 'svg', 'tiff', 'jpg', 'jpeg'}
    if file_format not in allowed_formats:
        raise ValueError(f"Invalid format: {format}. Choose from {allowed_formats}")

    if save_path is not None:
        save_path = Path(save_path)
        save_path.mkdir(parents=True, exist_ok=True)

    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)

    if title is None:
        title = color if isinstance(color, str) else ', '.join(color)

    try:
        sc.pl.embedding(
            adata,
            basis='spatial',
            color=color,
            title=title,
            s=s,
            ax=ax,
            show=False,
            frameon=frameon,
            legend_loc=legend_loc,
            colorbar_loc=colorbar_loc, 
            **kwargs
        )
    except KeyError as e:
        raise ValueError(f"Missing required data: {e}") from None

    if save_path is not None:
        output_path = save_path / f"{save_name}.{file_format}"
        try:
            fig.savefig(
                output_path,
                dpi=dpi,
                bbox_inches='tight' if adjust_margins else None,
                pad_inches=0.1 if adjust_margins else 0.5
            )
        except Exception as e:
            raise IOError(f"Failed to save figure: {e}") from None

    if show:
        plt.show()
    
    plt.close(fig)

import pandas as pd
import numpy as np

def getLogFC(
    target_genes: list,
    target_groups: list,
    logfoldchanges: dict,
    gene_names: dict
) -> pd.DataFrame:
    r"""
    Get log fold changes for target genes and groups.

    Parameters:
        target_genes (list): List of target genes.
        target_groups (list): List of target groups.
        logfoldchanges (dict): Dictionary of log fold changes.
        gene_names (dict): Dictionary of gene names.

    Returns:
        pd.DataFrame: DataFrame containing log fold changes for target genes and groups.
    """
    
    if len(target_genes) != len(target_groups):
        raise ValueError("Lengths of `target_genes` and `target_groups` must match.")

    results = []
    for gene, group in zip(target_genes, target_groups):
        # Extract gene names for the current group
        group_genes = gene_names[group]
        # Find index of the target gene in the group's gene list
        gene_idx = np.where(group_genes == gene)[0]
        
        if len(gene_idx) == 0:
            print(f"Warning: Gene '{gene}' not found in group '{group}'.")
            results.append((gene, group, None))  # Store None if gene is missing
        else:
            # Get the first matching LogFC value (assumes unique gene names per group)
            logfc = logfoldchanges[group][gene_idx[0]]
            results.append((gene, group, logfc))
    
    # Convert results to a structured DataFrame
    results_df = pd.DataFrame(results, columns=["Gene", "Group", "LogFC"])
    return results_df


def rank_genes_groups(
    adata,
    groupby: str = "DePass",       
    method: str = "wilcoxon",        
    n_genes: int = 10,             
    standard_scale: str = "var",    
    dpi: int = 300,                
    show: bool = True, 
    save_path: Optional[str] = None,                
    figname: str = 'rank_genes_dotplot',  
) -> None:
    r"""
    Rank genes by groups and optionally plot the results.

    Parameters:
        adata (AnnData): Input AnnData object.
        groupby (str): Column name in adata.obs for grouping (default: "DePass").
        method (str): Method for ranking genes (default: "wilcoxon").
        n_genes (int): Number of top genes to show (default: 10).
        standard_scale (str): Scaling method (default: "var").
        dpi (int): DPI of the saved figure (default: 300).
        show (bool): Whether to show the results (default: True).
        save_path (Optional[str]): Path to save the figure (default: None).
        figname (str): Name of the saved figure (default: 'rank_genes_dotplot').

    Returns:
        None
    """
    
    from sklearn.preprocessing import MinMaxScaler
    import matplotlib.pyplot as plt
     
    if save_path is not None:
        os.makedirs(save_path, exist_ok=True)

    # Normalize data to [0,1] range for comparative analysis
    scaler = MinMaxScaler()
    adata.obs[groupby] = adata.obs[groupby].astype('str').astype('category')  # Ensure categorical type
    adata.X = scaler.fit_transform(adata.X) 
    
    # Perform differential expression analysis
    sc.tl.rank_genes_groups(adata, groupby=groupby, method=method, use_raw=False)

    if show or save_path is not None:
        sc.pl.rank_genes_groups_dotplot(
            adata,
            groupby=groupby,
            standard_scale=standard_scale,  
            n_genes=n_genes,
            show=False ,
            dendrogram=False,
        )
        if save_path is not None:
           plt.savefig(
               os.path.join(save_path, figname+".pdf"),
               dpi=dpi,
               bbox_inches="tight"  
           )

        if show: plt.show()
        plt.close()

 
from typing import Optional
import numpy as np
import matplotlib.pyplot as plt
import scanpy as sc
from sklearn.preprocessing import MinMaxScaler
import os
import warnings

def get_logfc(
    target_gene: str,
    target_group: str,
    logfoldchanges: np.ndarray,  # Structured array from rank_genes_groups
    gene_names: np.ndarray,      # Structured array from rank_genes_groups
) -> float:
   
    # Validate group existence
    if target_group not in gene_names.dtype.names:
        available_groups = list(gene_names.dtype.names)
        raise KeyError(f"Group '{target_group}' not found. Available groups: {available_groups}")
    
    # Locate gene index
    group_genes = gene_names[target_group]
    gene_idx = np.flatnonzero(group_genes == target_gene)
    
    if not gene_idx.size:
        raise KeyError(f"Gene '{target_gene}' not found in group '{target_group}'")
        
    return float(logfoldchanges[target_group][gene_idx[0]])


def plot_marker(
    adata: sc.AnnData,
    target_gene: str,
    save_path: Optional[str] = None,
    save_name: str = "",
    show: bool = True,
    s: int = 80,
    cmap: str = "viridis",
    dpi: int = 300,
    colorbar_loc: Optional[str] = None, 
    figsize: tuple = (3, 3),
    frameon=False,
) -> None:
    r"""
    Plot marker comparison between two datasets.

    Parameters:
        adata (AnnData): AnnData object.
        target_gene (str): Target gene name.
        save_path (Optional[str]): Path to save the figure (default: None).
        save_name (str): Name of the saved figure (default: "gene_comparison").
        show (bool): Whether to show the plot (default: True).
        s (int): Size of the markers (default: 80).
        cmap (str): Colormap (default: "viridis").
        dpi (int): DPI of the saved figure (default: 300).
        colorbar_loc (Optional[str]): Colorbar position. Set to None to disable colorbar.
        figsize (tuple): Figure size (default: (7, 3)).
        frameon (bool): Whether to show the frame (default: False).

    Returns:
        None
    """
    
    if 'spatial' not in adata.obsm:
            raise KeyError(f"Missing spatial coordinates in obsm['spatial']")
    
    if save_path is not None:
        os.makedirs(save_path, exist_ok=True)
    
    # Shared visualization parameters
    vis_params = {
        'basis': 'spatial',
        'color': f'{target_gene}_expr',
        's': s,
        'frameon': frameon,
        'colorbar_loc': colorbar_loc,
        'cmap': cmap,
    }

    def _scaler_data(adata: sc.AnnData) -> None:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            expr = adata[:, target_gene].X.toarray()
        adata.obs[f'{target_gene}_expr'] = MinMaxScaler().fit_transform(expr)

    def _create_plot(adata: sc.AnnData, ax: plt.Axes) -> None:
        sc.pl.embedding(
            adata,
            title=f"{target_gene}",
            ax=ax,
            show=False,
            **vis_params
        )

    
    _scaler_data(adata)
    
    fig, axes = plt.subplots(1, 1, figsize=figsize)
    _create_plot(adata, axes)
    
    if save_path is not None:
        fig.savefig(
            os.path.join(save_path, f"{save_name}.pdf"),
            dpi=dpi,
            bbox_inches="tight"
        )
    if show:
        plt.show()
    plt.close(fig)


def plot_marker_comparison(
    adata1: sc.AnnData,
    adata2: sc.AnnData,
    target_gene: str,
    save_path: Optional[str] = None,
    save_name: str = "gene_comparison",
    show: bool = False,
    s: int = 80,
    cmap: str = "viridis",
    dpi: int = 300,
    colorbar_loc: Optional[str] = None, 
    figsize: tuple = (7, 3),
    frameon=False,
) -> None:
    r"""
    Plot marker comparison between two datasets.

    Parameters:
        adata1 (AnnData): First AnnData object.
        adata2 (AnnData): Second AnnData object.
        target_gene (str): Target gene name.
        save_path (Optional[str]): Path to save the figure (default: None).
        save_name (str): Name of the saved figure (default: "gene_comparison").
        show (bool): Whether to show the plot (default: False).
        s (int): Size of the markers (default: 80).
        cmap (str): Colormap (default: "viridis").
        dpi (int): DPI of the saved figure (default: 300).
        colorbar_loc (Optional[str]): Colorbar position. Set to None to disable colorbar.
        figsize (tuple): Figure size (default: (7, 3)).
        frameon (bool): Whether to show the frame (default: False).

    Returns:
        None
    """
    
    for adata, name in [(adata1, 'adata1'), (adata2, 'adata2')]:
        if 'spatial' not in adata.obsm:
            raise KeyError(f"Missing spatial coordinates in {name}.obsm['spatial']")
    
    if save_path is not None:
        os.makedirs(save_path, exist_ok=True)
    
    vis_params = {
        'basis': 'spatial',
        'color': f'{target_gene}_expr',
        's': s,
        'frameon':   frameon,
        'colorbar_loc': colorbar_loc,
        'cmap': cmap,
    }

    def _scaler_data(adata: sc.AnnData) -> None:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            expr = adata[:, target_gene].X.toarray()
        adata.obs[f'{target_gene}_expr'] = MinMaxScaler().fit_transform(expr)

    def _create_plot(adata: sc.AnnData, ax: plt.Axes) -> None:
        sc.pl.embedding(
            adata,
            title=f"{target_gene}",
            ax=ax,
            show=False,
            **vis_params
        )

    
    _scaler_data(adata1)
    _scaler_data(adata2)
    
    fig, axes = plt.subplots(1, 2, figsize=figsize)
    _create_plot(adata1, axes[0])
    _create_plot(adata2, axes[1])
    
    if save_path is not None:
        # Save combined comparison
        fig.savefig(
            os.path.join(save_path, f"{save_name}_combined.pdf"),
            dpi=dpi,
            bbox_inches="tight"
        )
    if show:
        plt.show()
    plt.close(fig)


def plot_marker_comparison_with_logFC(
    adata1: sc.AnnData,
    adata2: sc.AnnData,
    target_gene: str,
    target_group: str,
    save_path: Optional[str] = None,
    save_name: str = "gene_comparison",
    show: bool = False,
    s: int = 80,
    cmap: str = "viridis",
    dpi: int = 300,
    colorbar_loc: Optional[str] = None, 
    figsize: tuple = (7, 3),
    frameon=False,
) -> None:
    r"""
    Visualize marker comparison between two datasets with log fold changes.

    Parameters:
        adata1 (AnnData): First AnnData object.
        adata2 (AnnData): Second AnnData object.
        target_gene (str): Target gene name.
        target_group (str): Target group name.
        save_path (Optional[str]): Path to save the figure (default: None).
        save_name (str): Name of the saved figure (default: "gene_comparison").
        show (bool): Whether to show the plot (default: False).
        s (int): Size of the markers (default: 80).
        cmap (str): Colormap (default: "viridis").
        dpi (int): DPI of the saved figure (default: 300).
        colorbar_loc (Optional[str]): Colorbar position. Set to None to disable colorbar.
        figsize (tuple): Figure size (default: (7, 3)).
        frameon (bool): Whether to show the frame (default: False).

    Returns:
        None
    """
    
    for adata, name in [(adata1, 'adata1'), (adata2, 'adata2')]:
        if 'rank_genes_groups' not in adata.uns:
            raise KeyError(f"Missing DEG results in {name}. Run sc.tl.rank_genes_groups first.")
        if 'spatial' not in adata.obsm:
            raise KeyError(f"Missing spatial coordinates in {name}.obsm['spatial']")
   
    if save_path is not None:
        os.makedirs(save_path, exist_ok=True)
    
    logfc1 = get_logfc(target_gene, target_group,
                               adata1.uns['rank_genes_groups']['logfoldchanges'],
                               adata1.uns['rank_genes_groups']['names'])
    
    logfc2 = get_logfc(target_gene, target_group,
                               adata2.uns['rank_genes_groups']['logfoldchanges'],
                               adata2.uns['rank_genes_groups']['names'])
    
    if save_path is not None:
        os.makedirs(save_path, exist_ok=True)
    
    # Shared visualization parameters
    vis_params = {
        'basis': 'spatial',
        'color': f'{target_gene}_expr',
        's': s,
        'frameon': frameon,
        'colorbar_loc': colorbar_loc,
        'cmap': cmap,
    }

    def _scaler_data(adata: sc.AnnData) -> None:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            expr = adata[:, target_gene].X.toarray()
        adata.obs[f'{target_gene}_expr'] = MinMaxScaler().fit_transform(expr)

    def _create_plot(adata: sc.AnnData, logfc: float, ax: plt.Axes, name:str) -> None:
        sc.pl.embedding(
            adata,
            title=f"{name+target_gene}\n(logFC={logfc:.3f})",
            ax=ax,
            show=False,
            **vis_params
        )

    for adata in [adata1, adata2]:
        _scaler_data(adata)
    
    fig, axes = plt.subplots(1, 2, figsize=figsize)
    _create_plot(adata1, logfc1, axes[0],'Raw-')
    _create_plot(adata2, logfc2, axes[1],'Enhanced-')
    
    if save_path is not None:
        fig.savefig(
            os.path.join(save_path, f"{save_name}_combined_logFC.pdf"),
            dpi=dpi,
            bbox_inches="tight"
        )
    if show:
        plt.show()
    plt.close(fig)


#####  liwei shared the ways of visualized  ####
"""
(1) final_embeddings is a dictionary, and you can input {'s1': embedding_array}.
(2) data_dict is a data dictionary, and you can input:
python
data_dict = {  
    'RNA': [adata1_rna],  
    'Protein': [adata1_adt],  
}  
(3) This function supports K-means clustering by default, so you can specify n_clusters.
(4) If you have already obtained clusters using mclust, set mode='defined' and pass your cluster labels in defined_label (should be an array).
(5) vis_basis refers to the spatial coordinates stored in adata, typically in obsm['spatial']. Note: It is recommended to divide all coordinates 
by 20 to convert them into a grid format (e.g., 0, 1, 2, 3, 4) rather than keeping them at intervals of 20.

"""
from sklearn.cluster import KMeans
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
import os
import matplotlib.cm as cm
from matplotlib.colors import to_rgb
import matplotlib.patches as patches

# cluster visualization
def cluster_and_visualize_superpixel(
    final_embeddings,
    data_dict,
    n_clusters,
    mode="joint",  # 'joint' or 'independent'
    defined_labels=None,
    vis_basis="spatial",
    random_state=0,
    colormap=None,
    swap_xy=False,
    invert_x=False,
    invert_y=False,
    offset=False,
    save_path=None,
    dpi=300,
    remove_title = False,
    remove_legend = False,
    remove_spine = False,
    figscale = 35
):
    import numpy as np
    from sklearn.cluster import KMeans
    import os
    import numpy as np


    adata_list = []
    embeddings = []
    coords_all = []
    section_names = []

    for section, embedding in final_embeddings.items():
        idx = int(section[1:]) - 1
        for modality, adata_list_per_mod in data_dict.items():
            if idx < len(adata_list_per_mod) and adata_list_per_mod[idx] is not None:
                adata = adata_list_per_mod[idx]
                adata_list.append(adata)
                embeddings.append(embedding)
                coords = adata.obsm[vis_basis].copy()
                if swap_xy:
                    coords = coords[:, [1, 0]]
                coords = coords.astype(int)
                if offset:
                    offset_value = coords.min(axis=0)     
                    coords -= offset_value               
                coords_all.append(coords)
                section_names.append(section)
                break

    cluster_labels = {}

    if mode == "joint":
        print("Perform joint clustering...")
        combined_embedding = np.vstack(embeddings)
        kmeans = KMeans(n_clusters=n_clusters, random_state=random_state)
        all_clusters = kmeans.fit_predict(combined_embedding)
        start = 0
        for section, emb in zip(section_names, embeddings):
            end = start + emb.shape[0]
            cluster_labels[section] = all_clusters[start:end]
            start = end
    elif mode == "independent":
        print("Perform independent clustering...")
        for section, emb in zip(section_names, embeddings):
            kmeans = KMeans(n_clusters=n_clusters, random_state=random_state)
            cluster_labels[section] = kmeans.fit_predict(emb)
    elif mode == 'defined':
        if defined_labels is None:
            raise ValueError("If mode='defined', you must provide `defined_labels`.")
        cluster_labels = defined_labels
    else:
        raise ValueError("mode must be 'joint' or 'independent'")

    for section, coords, labels in zip(section_names, coords_all, cluster_labels.values()):
        max_y, max_x = coords.max(axis=0) + 1
        image = np.full((max_y, max_x), fill_value=-1, dtype=int)
        for (y, x), label in zip(coords, labels):
            image[y, x] = label
        if invert_x:
            image = image[:, ::-1]
        if invert_y:
            image = image[::-1, :]
        section_save_path = None
        if save_path:
            base, ext = os.path.splitext(save_path)
            section_save_path = f"{base}_section_{section}{ext or '.png'}"
        
        plot_histology_clusters(
            he_clusters_image=image,
            num_he_clusters=n_clusters,
            section_title=f"Section {section} ({mode})",
            colormap=colormap,
            save_path=section_save_path,
            dpi=dpi,
            figscale = figscale,
            remove_title = remove_title,
            remove_legend = remove_legend,
            remove_spine=remove_legend, 
        )

    return cluster_labels



def plot_histology_clusters(he_clusters_image,
                            num_he_clusters,
                            section_title=None,
                            colormap=None,
                            save_path=None,
                            figscale = 35,
                            remove_title = False,
                            remove_legend = False,
                            remove_spine=False, 
                            dpi=300):


    if colormap is None:
        color_list = [[255,127,14],[44,160,44],[214,39,40],[148,103,189],
                      [140,86,75],[227,119,194],[127,127,127],[188,189,34],
                      [23,190,207],[174,199,232],[255,187,120],[152,223,138],
                      [255,152,150],[197,176,213],[196,156,148],[247,182,210],
                      [199,199,199],[219,219,141],[158,218,229],[16,60,90],
                      [128,64,7],[22,80,22],[107,20,20],[74,52,94],[70,43,38],
                      [114,60,97],[64,64,64],[94,94,17],[12,95,104],[0,0,0]]

    elif isinstance(colormap, list):
        color_list = colormap

    else:
        cmap = cm.get_cmap(colormap)
        color_list = [ [int(255 * c) for c in to_rgb(cmap(i))] for i in range(len(cmap.colors)) ]

    image_rgb = 255 * np.ones([he_clusters_image.shape[0], he_clusters_image.shape[1], 3])
    for cluster in range(num_he_clusters):
        image_rgb[he_clusters_image == cluster] = color_list[cluster]
    image_rgb = np.array(image_rgb, dtype='uint8')

    plt.figure(figsize=(he_clusters_image.shape[1] // figscale, he_clusters_image.shape[0] // figscale))
    if remove_title:
        plt.title("")
    else:
        title = section_title if section_title else "Histology Clusters"
        plt.title(title, fontsize=18)
    plt.imshow(image_rgb, interpolation='none')
    # plt.show()
    ax = plt.gca()
    ax.set_xticks([])
    ax.set_yticks([])

    if remove_spine:
        for spine in ax.spines.values():
            spine.set_visible(False)

    if not remove_legend:
        legend_elements = [patches.Patch(facecolor=np.array(color_list[i]) / 255,
                                         label=f'Cluster {i}')
                           for i in range(num_he_clusters)]
        plt.legend(handles=legend_elements,
                   bbox_to_anchor=(1.05, 1),
                   loc='upper left',
                   borderaxespad=0.,
                   fontsize=12)

    if save_path is not None:
        plt.savefig(save_path, dpi=dpi, bbox_inches="tight")
        print(f"Saved: {save_path}")

    plt.show()
    # plt.close()

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
import matplotlib.cm as cm
from matplotlib.colors import to_rgb
import matplotlib.patches as patches
import os


def plot_superpixel(
    adata,
    label_key='label',  
    vis_basis='spatial',  
    colormap=None,
    save_path=None,
    save_name='cluster_visualization',
    title=None,
    figscale=100,
    format='png',
    show=True,
    remove_title=False,
    remove_legend=False,
    remove_spine=False,
    dpi=300,
    random_state=2024,  
    swap_xy=False,  
    invert_x=False,  
    invert_y=False  
):
    """
    Visualize clusters using labels and spatial coordinates from adata.

    Parameters:
        adata: AnnData object containing cell data and cluster labels.
        label_key: Key for cluster labels in adata.obs.
        vis_basis: Key for coordinates in adata.obsm.
        colormap: Custom color mapping.
        save_path: Directory path to save the image.
        save_name: File name to save the image.
        title: Title of the plot.
        figscale: Figure size scaling factor.
        format: Format to save the image (e.g., 'pdf', 'png').
        show: Whether to display the plot.
        remove_title: Whether to remove the title.
        remove_legend: Whether to remove the legend.
        remove_spine: Whether to remove the border.
        dpi: Image resolution.
        random_state: Random seed for reproducibility.
        swap_xy: Whether to swap x and y coordinates.
        invert_x: Whether to invert x-axis.
        invert_y: Whether to invert y-axis.
    """
   
    np.random.seed(random_state)

   
    labels = adata.obs[label_key].values.astype(int)
    coords = adata.obsm[vis_basis].copy().astype(int)

    if swap_xy:
        coords = coords[:, [1, 0]]

    unique_labels = np.unique(labels)
    num_clusters = len(unique_labels)

    label_to_index = {label: idx for idx, label in enumerate(unique_labels)}
    zero_based_labels = np.array([label_to_index[label] for label in labels])


    if colormap is None:
        color_list = [
    [60, 142, 204],    
    [187, 187, 187],   
    [246, 216, 208],   
    [254, 238, 237],   
    [215, 102, 102],   
    [177, 157, 177],   
    [60, 162, 254],    
    [151, 215, 243],   
    [208, 163, 239],   
    [246, 216, 212],   
    [255, 247, 180],   
    [241, 91, 108],    
    [60, 188, 60],     
    [104, 220, 104],   
    [247, 172, 188],   
    [222, 171, 138],   
    [255, 188, 188],   
    [199, 133, 89],    
    [60, 251, 255],    
    [195, 236, 255],   
    [204, 238, 204],   
    [254, 220, 189],   
    [239, 91, 156],    
    [176,224,230],   
    [187, 187, 187]    
]
    elif isinstance(colormap, list):
        color_list = colormap
    else:
        cmap = cm.get_cmap(colormap)
        color_list = [[int(255 * c) for c in to_rgb(cmap(i))] for i in range(len(cmap.colors))]

    if len(color_list) < num_clusters:
        raise ValueError("Color list is not long enough to cover all clusters.")

    max_y, max_x = coords.max(axis=0) + 1
    image = np.full((max_y, max_x), fill_value=-1, dtype=int)
    for (y, x), label in zip(coords, zero_based_labels):
        if 0 <= x < max_x and 0 <= y < max_y:
            image[y, x] = label

    if invert_x:
        image = image[:, ::-1]
    if invert_y:
        image = image[::-1, :]

    image_rgb = 255 * np.ones([image.shape[0], image.shape[1], 3])
    for cluster in range(num_clusters):
        image_rgb[image == cluster] = color_list[cluster]
    image_rgb = np.array(image_rgb, dtype='uint8')

    
    plt.figure(figsize=(image.shape[1] // figscale, image.shape[0] // figscale))
    if remove_title or title is None:
        plt.title("")
    else:
        plt.title(title, fontsize=18)
    plt.imshow(image_rgb, interpolation='none')
    ax = plt.gca()
    ax.set_xticks([])
    ax.set_yticks([])

    if remove_spine:
        for spine in ax.spines.values():
            spine.set_visible(False)

    if not remove_legend:
        legend_elements = [patches.Patch(facecolor=np.array(color_list[i]) / 255,
                                          label=f'Cluster {unique_labels[i]}') for i in range(num_clusters)]
        plt.legend(handles=legend_elements,
                   bbox_to_anchor=(1.05, 1),
                   loc='upper left',
                   borderaxespad=0.,
                   fontsize=12)

    if save_path:
        os.makedirs(save_path, exist_ok=True)
        file_path = os.path.join(save_path, f"{save_name}.{format}")
        plt.savefig(file_path, dpi=dpi, bbox_inches="tight", format=format)
        print(f"Image saved to: {file_path}")

    if show:
        plt.show()
    else:
        plt.close()


# marker visualization
def prepare_image(adata, molecule_name, basis, swap_xy, invert_x, invert_y, offset, scale):
    coords = adata.obsm[basis].copy()
    if swap_xy:
        coords = coords[:, [1, 0]]
    coords = coords.astype(int)
    if offset:
        offset_value = coords.min(axis=0)
        coords -= offset_value 

    values = adata[:, molecule_name].X
  

    if hasattr(values, "toarray"):
        values = values.toarray().flatten()
    else:
        values = np.array(values).flatten()

    if scale:
        from sklearn.preprocessing import MinMaxScaler
        scaler = MinMaxScaler()
        values = values.reshape(-1, 1)  
        values = scaler.fit_transform(values)
        values = values.flatten()  
  

    max_y, max_x = coords.max(axis=0) + 1
    image = np.full((max_y, max_x), np.nan, dtype=float)
    for (y, x), val in zip(coords, values):
        image[y, x] = val

    if invert_x:
        image = image[:, ::-1]
    if invert_y:
        image = image[::-1, :]

    return image

def plot_marker_comparison_superpixel(
    molecule_name: str,
    adata1,
    adata2,
    section1_label: str = 'Section 1',
    section2_label: str = 'Section 2',
    basis: str = 'spatial',
    colormap: str = "viridis",
    plot_style: str = "original",
    scale: bool = True,
    swap_xy: bool = False,
    invert_x: bool = False,
    invert_y: bool = False,
    offset: bool = False,
    figscale: int = 35,
    dpi: int = 300,
    remove_title: bool = False,     
    remove_spine: bool = False,    
    remove_legend: bool = False,      
    save_path: str = None
):


    img1 = prepare_image(adata1, molecule_name, basis, swap_xy, invert_x, invert_y, offset, scale)
    img2 = prepare_image(adata2, molecule_name, basis, swap_xy, invert_x, invert_y, offset, scale)


    figsize1 = (img1.shape[1] / figscale, img1.shape[0] / figscale)
    figsize2 = (img2.shape[1] / figscale, img2.shape[0] / figscale)
    figsize = (figsize1[0] + figsize2[0], max(figsize1[1], figsize2[1]))

    fig, axes = plt.subplots(1, 2, figsize=figsize)
    for ax, img, title in zip(axes, [img1, img2], [section1_label, section2_label]):
        im = ax.imshow(img, cmap=colormap, interpolation='none')
        if not remove_title:
            ax.set_title(f"{title} - {molecule_name}", fontsize=16)
        else:
            ax.set_title("")
        ax.set_xticks([])
        ax.set_yticks([])
        if remove_spine:
            for spine in ax.spines.values():
                spine.set_visible(False)
        if plot_style == "equal":
            ax.set_aspect("equal")

        if not remove_legend:
            cbar = fig.colorbar(im, ax=ax, shrink=0.7, pad=0.02)  

    if save_path:
        base, ext = os.path.splitext(save_path)
        if not ext:
            ext = ".png"
        save_path = base + ext
        os.makedirs(os.path.dirname(save_path), exist_ok=True)
        print(f"Saving marker comparison to: {save_path}")
        plt.savefig(save_path, dpi=dpi, bbox_inches='tight')

    plt.show()
    plt.close()


import pandas as pd
from typing import List, Dict, Union
from pathlib import Path
from anndata import AnnData


def get_logfc_df(
    adata_list: List[AnnData],
    adata_names: List[str],
    target_genes: List[str],
    target_groups: List[str],
    save_path: Union[str, Path] = "results",
    save_name: str = "logfc_comparison"
) -> pd.DataFrame:
    r"""
    Get log fold change DataFrame for multiple datasets.

    Parameters:
        adata_list (List[AnnData]): List of AnnData objects.
        adata_names (List[str]): List of dataset names.
        target_genes (List[str]): List of target genes.
        target_groups (List[str]): List of target groups.
        save_path (Union[str, Path]): Path to save the DataFrame (default: "results").
        save_name (str): Name of the saved file (default: "logfc_comparison").

    Returns:
        pd.DataFrame: DataFrame containing log fold changes.
    """
    
    if len(target_genes) != len(target_groups):
        raise ValueError("Length of target_genes and target_groups must match!")
    
    if len(adata_list) != len(adata_names):
        raise ValueError("Length of adata_list and adata_names must match!")
    
    results = {
        'Gene': target_genes,
        'Group': target_groups
    }
    
    # Extract logFC values for each dataset
    for adata, name in zip(adata_list, adata_names):
        logfc_values = []
        for gene, group in zip(target_genes, target_groups):
            try:
                logfc = get_logfc(
                    target_gene=gene,
                    target_group=group,
                    logfoldchanges=adata.uns['rank_genes_groups']['logfoldchanges'],
                    gene_names=adata.uns['rank_genes_groups']['names']
                )
                logfc_values.append(logfc)
            except KeyError as e:
                print(f"[Warning] Failed to retrieve {gene}@{group} in dataset {name}: {str(e)}")
                logfc_values.append(None)
        
        results[f'logFC_{name}'] = logfc_values
    
    # Convert to DataFrame
    logfc_df = pd.DataFrame(results)
    
    # Melt the DataFrame for visualization
    value_vars = [col for col in logfc_df.columns if col.startswith('logFC_')]
    long_df = pd.melt(
        logfc_df,
        id_vars=['Gene', 'Group'],
        value_vars=value_vars,
        var_name='type',
        value_name='logFC'
    )
    
    long_df['type'] = long_df['type'].str.replace('logFC_', '')
    
    output_path = Path(save_path)
    output_path.mkdir(parents=True, exist_ok=True)
    long_df.to_csv(output_path / f"{save_name}.csv", sep='\t', index=False)
    
    return long_df

import pandas as pd
from anndata import AnnData
from typing import Optional, Union

def get_top_degs_df(
    adata: AnnData,
    n_top_genes: int = 20,
    groupby: Optional[str] = None,
    key: str = 'rank_genes_groups'
) -> pd.DataFrame:
    r"""
    Extract top N differentially expressed genes (DEGs) and their statistics from Scanpy's rank_genes_groups results.
    
    Parameters
    ----------
    adata : AnnData
        AnnData object containing rank_genes_groups results.
    n_top_genes : int, optional
        Number of top DEGs to extract per group (default: 20).
    groupby : str, optional
        Column name in adata.obs used for grouping. If None, automatically reads from rank_genes_groups.
    key : str, optional
        Key name for rank_genes_groups in adata.uns (default: 'rank_genes_groups').
    
    Returns
    -------
    pd.DataFrame
        DataFrame containing the following columns:
        - Group: Group name
        - Gene: Gene symbol
        - LogFC: Log fold change
        - PValue: Raw p-value
        - AdjPValue: Adjusted p-value (e.g., FDR)
    """

    if key not in adata.uns:
        raise KeyError(f"'{key}' not found in adata.uns. Run sc.tl.rank_genes_groups first.")
    
    rank_data = adata.uns[key]
    required_fields = ['names', 'logfoldchanges', 'pvals', 'pvals_adj']
    for field in required_fields:
        if field not in rank_data:
            raise ValueError(f"Missing required field '{field}' in rank_genes_groups data.")

    if groupby is None:
        groupby = rank_data['params']['groupby'] if 'params' in rank_data else None
        if groupby is None:
            raise ValueError("Please specify them manually through the 'groupby' parameter.")


    groups = rank_data['names'].dtype.names
    gene_names = rank_data['names']
    logfcs = rank_data['logfoldchanges']
    pvals = rank_data['pvals']
    pvals_adj = rank_data['pvals_adj']

    top_genes = []
    for group in groups:
        genes = gene_names[group][:n_top_genes]
        valid_idx = ~pd.isnull(genes)
        genes = genes[valid_idx]
        
        group_logfcs = logfcs[group][:n_top_genes][valid_idx]
        group_pvals = pvals[group][:n_top_genes][valid_idx]
        group_padjs = pvals_adj[group][:n_top_genes][valid_idx]
        
        for gene, lfc, pval, padj in zip(genes, group_logfcs, group_pvals, group_padjs):
            top_genes.append({
                'Group': group,
                'Gene': gene,
                'LogFC': lfc,
                'PValue': pval,
                'AdjPValue': padj
            })

    return pd.DataFrame(top_genes)

from typing import Optional, Dict
import os
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from anndata import AnnData

def plot_modality_weights(
    adata: AnnData,
    modality_names: tuple = ("RNA", "Protein"),
    cluster_column: str = "DePass",
    save_path: Optional[str] = None,
    save_name: str = "modality_weights",
    show: bool = True,
    figsize: tuple = (5, 3),
    palette: Dict[str, str] = None,
    **kwargs
) -> plt.Axes:
    r"""
    Plot modality weights for clusters with legend outside right.
    """
 
    if 'alpha' not in adata.obsm:
        raise KeyError("Missing modality weights in adata.obsm['alpha']")
    if cluster_column not in adata.obs:
        raise KeyError(f"Cluster column '{cluster_column}' not found in adata.obs")
    if adata.obsm['alpha'].shape[1] != 2:
        raise ValueError(f"Expected 2 columns in adata.obsm['alpha'], found {adata.obsm['alpha'].shape[1]}")

    default_palette = {modality_names[0]: "#CAB82E", modality_names[1]: "#9368A6"}
    palette = palette or default_palette

    plot_df = pd.DataFrame({
        modality_names[0]: adata.obsm['alpha'][:, 0],
        modality_names[1]: adata.obsm['alpha'][:, 1],
        'Cluster': adata.obs[cluster_column].astype(str)
    })
    melted_df = plot_df.melt(
        id_vars='Cluster',
        value_vars=modality_names,
        var_name='Modality',
        value_name='Weight'
    )
    clusters = sorted(plot_df['Cluster'].unique(), key=lambda x: int(x))
    melted_df['Cluster'] = pd.Categorical(melted_df['Cluster'], categories=clusters, ordered=True)

    plt.figure(figsize=figsize)
    ax = sns.violinplot(
        data=melted_df,
        x='Cluster',
        y='Weight',
        hue='Modality',
        inner="quart",
        linewidth=0.5,
        palette=palette,
        **kwargs
    )

    ax.set_title(f"{modality_names[0]} vs {modality_names[1]}", pad=15)
    ax.set_xlabel("Cluster", labelpad=10)
    ax.set_ylabel("Attention Weight", labelpad=10)

    ax.legend(
        loc='center left',
        bbox_to_anchor=(1.05, 0.5),
        frameon=True,
        title='Modality'
    )
    
    plt.subplots_adjust(right=0.75)
    plt.tight_layout()

    if save_path:
        os.makedirs(save_path, exist_ok=True)
        plt.savefig(
            os.path.join(save_path, f"{save_name}.pdf"),
            dpi=300,
            bbox_inches="tight",
            transparent=True
        )
    if show:
        plt.show()
    plt.close()

    return ax


def noise(adata_omics1, adata_omics2, level):
    r"""
    Add noise to omics data.

    Parameters:
        adata_omics1 (AnnData): First AnnData object.
        adata_omics2 (AnnData): Second AnnData object.
        level (float): Noise level.

    Returns:
        tuple: Noisy omics data.
    """
    from scipy import stats
    noise1 = stats.norm.rvs(loc=0, scale=level, size=adata_omics1.obsm['X_norm'].shape)
    noise2 = stats.norm.rvs(loc=0, scale=level, size=adata_omics2.obsm['X_clr'].shape)

    return adata_omics1.obsm['X_norm'] + noise1, adata_omics2.obsm['X_clr'] + noise2


def Generate_masked_data(path,mask_ratio):
    r"""
    Generate masked data.

    Parameters:
        mask_ratio (float): Masking ratio.

    Returns:
        tuple: Masked data.
    """
    import numpy as np
    from scipy import sparse
    import scanpy as sc
    from DePass.utils import preprocess_data

    adata_omics1 = sc.read_h5ad(path+'/adata_RNA.h5ad')
    adata_omics2 = sc.read_h5ad(path+'/adata_ADT.h5ad')  
    
    adata_omics1.var_names_make_unique()
    adata_omics2.var_names_make_unique()

    def mask(adata, ratio):
        X = adata.X.copy()  

        if sparse.issparse(X):
            nonzero_positions = X.nonzero()
        else:
            nonzero_positions = np.nonzero(X)

        num_to_mask = int(ratio * len(nonzero_positions[0]))
        mask_indices = np.random.choice(len(nonzero_positions[0]), num_to_mask, replace=False)
        
        if sparse.issparse(X):
            X[(nonzero_positions[0][mask_indices], nonzero_positions[1][mask_indices])] = 0
        else:
            X[nonzero_positions[0][mask_indices], nonzero_positions[1][mask_indices]] = 0

        adata.X = X
        return adata

    sc.pp.filter_genes(adata_omics1, min_cells=adata_omics1.shape[0] * 0.01)
    sc.pp.highly_variable_genes(adata_omics1, flavor="seurat_v3", n_top_genes=1000)
    adata_omics1 = adata_omics1[:, adata_omics1.var['highly_variable']]
    adata_omics1 = mask(adata_omics1, mask_ratio)
    sc.pp.normalize_total(adata_omics1, target_sum=1e4)
    sc.pp.log1p(adata_omics1)
    sc.pp.scale(adata_omics1)

    adata_omics2 = mask(adata_omics2, mask_ratio)
    preprocess_data(adata_omics2, modality='protein')
    return adata_omics1.X, adata_omics2.X


import scanpy as sc
import pandas as pd
import numpy as np
from scipy.stats import pearsonr
import matplotlib.pyplot as plt
import seaborn as sns

def calculate_correlation(adata1, adata2, gene_adt_mapping):
    r"""
    Calculate correlations between genes and ADTs.

    Parameters:
        adata1 (AnnData): First AnnData object (genes).
        adata2 (AnnData): Second AnnData object (ADTs).
        gene_adt_mapping (dict): Mapping from genes to ADTs.

    Returns:
        pd.DataFrame: DataFrame containing correlation results.
    """
    results = []
    for adt, genes in gene_adt_mapping.items():
        if adt not in adata2.var_names:
            print(f"Warning: {adt} not found in ADT data.")
            continue
        
        for gene in genes:
            if gene not in adata1.var_names:
                print(f"Warning: {gene} not found in gene data.")
                continue
            
            # Extract expression values
            gene_expression = adata1[:, gene].X.flatten()  # Gene expression
            adt_expression = adata2[:, adt].X.flatten()    # ADT expression
            
            # Calculate Pearson correlation coefficient
            correlation, p_value = pearsonr(gene_expression, adt_expression)
            
            results.append({
                "ADT": adt,
                "Gene": gene,
                "Gene_ADT": f"{gene}_{adt}",  
                "Correlation": correlation,
                "P_value": p_value
            })
    
    return pd.DataFrame(results)


from setuptools import setup, find_packages

setup(
    name='DePass',
    version='0.0.12',
    description='DePass: a dual-enhanced graph learning framework for paired data integration',
    packages=find_packages(),
    install_requires=[
        'anndata>=0.9.2',        
        'hnswlib>=0.8.0',
        'matplotlib>=3.7.1,<4.0.0',  
        'numpy>=1.23.5,<2.0.0',      
        'pandas>=1.5.3,<2.0.0',
        'scanpy>=1.9.8,<2.0.0',
        'scikit-learn>=1.3.0,<2.0.0',
        'scipy>=1.10.1,<2.0.0',
        'seaborn>=0.12.2,<0.14.0',
        'tqdm>=4.65.0,<5.0.0',
        'jupyter',                   
    ],
    python_requires='>=3.8,<3.9',   
)